from __future__ import annotations

import contextlib

from prettyqt import constants
from prettyqt.qt import QtGui
from prettyqt.utils import get_repr


class UserData(QtGui.QTextBlockUserData):
    def __init__(self, data):
        super().__init__()
        self.data = data

    def __repr__(self):
        return get_repr(self, self.data)


class TextBlock(QtGui.QTextBlock):
    def __repr__(self):
        return get_repr(self, self.text())

    def __contains__(self, position: int):
        return self.contains(position)

    def __bool__(self):
        return self.isValid()

    def __str__(self):
        return self.text()

    def get_previous(self) -> TextBlock:
        return TextBlock(self.previous())

    def get_next(self) -> TextBlock:
        return TextBlock(self.next())

    def get_text_direction(self) -> constants.LayoutDirectionStr:
        return constants.LAYOUT_DIRECTION.inverse[self.textDirection()]

    def set_user_data(self, data):
        if isinstance(data, QtGui.QTextBlockUserData):
            self.setUserData(data)
            return None
        user_data = UserData(data)
        self.setUserData(user_data)

    def get_user_data(self):
        user_data = self.userData()
        return user_data.data if isinstance(user_data, UserData) else user_data

    @contextlib.contextmanager
    def edit_user_state(self):
        state = self.userState()
        yield state
        self.setUserState(state)

    def get_state(self) -> int:
        """Get the user state, generally used for syntax highlighting.

        :return: The block state
        """
        state = self.userState()
        return state if state == -1 else state & 0x0000FFFF

    def set_state(self, state: int):
        """Set the user state, generally used for syntax highlighting.

        :param state: new state value.
        """
        user_state = self.userState()
        if user_state == -1:
            user_state = 0
        higher_part = user_state & 0x7FFF0000
        state &= 0x0000FFFF
        state |= higher_part
        self.setUserState(state)

    def get_fold_level(self) -> int:
        """Get the block fold level.

        :return: The block fold level
        """
        state = self.userState()
        if state == -1:
            state = 0
        return (state & 0x03FF0000) >> 16

    def set_fold_level(self, val: int):
        """Set the block fold level.

        :param val: The new fold level [0-7]
        """
        state = self.userState()
        if state == -1:
            state = 0
        val = min(val, 0x3FF)
        state &= 0x7C00FFFF
        state |= val << 16
        self.setUserState(state)

    def is_fold_trigger(self) -> bool:
        """Check if the block is a fold trigger.

        :return: True if the block is a fold trigger (represented as a node in
            the fold panel)
        """
        state = self.userState()
        if state == -1:
            state = 0
        return bool(state & 0x04000000)

    def set_fold_trigger(self, val: int):
        """Set the block fold trigger flag (True means the block is a fold trigger).

        :param val: value to set
        """
        state = self.userState()
        if state == -1:
            state = 0
        state &= 0x7BFFFFFF
        state |= val << 26
        self.setUserState(state)

    def is_collapsed(self) -> bool:
        """Check if the block is expanded or collased.

        :return: False for an open trigger, True for for closed trigger
        """
        state = self.userState()
        if state == -1:
            state = 0
        return bool(state & 0x08000000)

    def set_collapsed(self, val: int):
        """Set the fold trigger state (collapsed or expanded).

        :param val: The new trigger state (True=collapsed, False=expanded)
        """
        state = self.userState()
        if state == -1:
            state = 0
        state &= 0x77FFFFFF
        state |= val << 27
        self.setUserState(state)

    def find_parent_scope(self, limit: int = 5000) -> TextBlock | None:
        """Find parent scope, if the block is not a fold trigger."""
        # if we moved up for more than n lines, just give up otherwise this
        # would take too much time.
        counter = 0
        original = TextBlock(self)
        start = TextBlock(self)
        if not self.is_fold_trigger():
            # search level of next non blank line
            while start.text().strip() == "" and start.isValid():
                start = start.next()
            ref_level = self.get_fold_level() - 1
            start = original
            while (
                start.blockNumber()
                and counter < limit
                and (not self.is_fold_trigger() or self.get_fold_level() > ref_level)
            ):
                counter += 1
                start = start.previous()
        return TextBlock(start) if counter < limit else None


if __name__ == "__main__":
    from prettyqt import gui

    doc = gui.TextDocument()
    doc.set_text("a\nb\ncd")
    block = doc[0]
    block.set_user_data("testg")



from __future__ import annotations

from prettyqt import widgets


class FocusFrame(widgets.WidgetMixin, widgets.QFocusFrame):
    """Focus frame which can be outside of a widget's normal paintable area."""


if __name__ == "__main__":
    app = widgets.app()
    container = widgets.Splitter()
    widget = widgets.PlainTextEdit()
    widget2 = widgets.PlainTextEdit()
    container.add(widget)
    container.add(widget2)
    errorbox = FocusFrame(container)
    errorbox.setWidget(container)
    errorbox.show()
    container.show()
    app.exec()



from __future__ import annotations

from typing import Literal

from prettyqt import constants, core, widgets
from prettyqt.qt import QtGui
from prettyqt.utils import bidict


StandardPixmapStr = Literal[
    "titlebar_min_button=",
    "titlebar_menu_button",
    "titlebar_max_button",
    "titlebar_close_button",
    "titlebar_normal_button",
    "titlebar_shade_button",
    "titlebar_unshade_button",
    "titlebar_context_help_button",
    "information",
    "warning",
    "critical",
    "question",
    "desktop",
    "trash",
    "computer",
    "drive_fd",
    "drive_hd",
    "drive_cd",
    "drive_dvd",
    "drive_net",
    "dir_home",
    "dir_open",
    "dir_closed",
    "dir",
    "dir_link",
    "dir_link_open",
    "file",
    "file_link",
    "file_dialog_start",
    "file_dialog_end",
    "file_dialog_to_parent",
    "file_dialog_new_folder",
    "file_dialog_detailed_view",
    "file_dialog_info_view",
    "file_dialog_contents_view",
    "file_dialog_list_view",
    "file_dialog_back",
    "dockwidget_close_button",
    "toolbar_horizontal_extension",
    "toolbar_vertical_extension",
    "dialog_ok",
    "dialog_cancel",
    "dialog_help",
    "dialog_open",
    "dialog_save",
    "dialog_close",
    "dialog_apply",
    "dialor_reset",
    "dialog_discard",
    "dialog_yes",
    "dialog_no",
    "arrow_up",
    "arrow_down",
    "arrow_left",
    "arrow_right",
    "arrow_back",
    "arrow_forward",
    "command_link",
    "vista_shield",
    "browser_reload",
    "browser_stop",
    "media_play",
    "media_stop",
    "media_pause",
    "media_skip_forward",
    "media_skip_backward",
    "media_seek_forward",
    "media_seek_backward",
    "media_volume",
    "media_volume_muted",
    "lineedit_clear",
    "custom_base",
    "dialog_yes_to_all",
    "dialog_no_to_all",
    "dialog_save_all",
    "dialog_abort",
    "dialog_retry",
    "dialog_ignore",
    "restore_defaults",
    "tab_close",
]

mod = widgets.QStyle

STANDARD_PIXMAP: bidict[StandardPixmapStr, mod.StandardPixmap] = bidict(
    titlebar_min_button=mod.StandardPixmap.SP_TitleBarMinButton,
    titlebar_menu_button=mod.StandardPixmap.SP_TitleBarMenuButton,
    titlebar_max_button=mod.StandardPixmap.SP_TitleBarMaxButton,
    titlebar_close_button=mod.StandardPixmap.SP_TitleBarCloseButton,
    titlebar_normal_button=mod.StandardPixmap.SP_TitleBarNormalButton,
    titlebar_shade_button=mod.StandardPixmap.SP_TitleBarShadeButton,
    titlebar_unshade_button=mod.StandardPixmap.SP_TitleBarUnshadeButton,
    titlebar_context_help_button=mod.StandardPixmap.SP_TitleBarContextHelpButton,
    information=mod.StandardPixmap.SP_MessageBoxInformation,
    warning=mod.StandardPixmap.SP_MessageBoxWarning,
    critical=mod.StandardPixmap.SP_MessageBoxCritical,
    question=mod.StandardPixmap.SP_MessageBoxQuestion,
    desktop=mod.StandardPixmap.SP_DesktopIcon,
    trash=mod.StandardPixmap.SP_TrashIcon,
    computer=mod.StandardPixmap.SP_ComputerIcon,
    drive_fd=mod.StandardPixmap.SP_DriveFDIcon,
    drive_hd=mod.StandardPixmap.SP_DriveHDIcon,
    drive_cd=mod.StandardPixmap.SP_DriveCDIcon,
    drive_dvd=mod.StandardPixmap.SP_DriveDVDIcon,
    drive_net=mod.StandardPixmap.SP_DriveNetIcon,
    dir_home=mod.StandardPixmap.SP_DirHomeIcon,
    dir_open=mod.StandardPixmap.SP_DirOpenIcon,
    dir_closed=mod.StandardPixmap.SP_DirClosedIcon,
    dir=mod.StandardPixmap.SP_DirIcon,
    dir_link=mod.StandardPixmap.SP_DirLinkIcon,
    dir_link_open=mod.StandardPixmap.SP_DirLinkOpenIcon,
    file=mod.StandardPixmap.SP_FileIcon,
    file_link=mod.StandardPixmap.SP_FileLinkIcon,
    file_dialog_start=mod.StandardPixmap.SP_FileDialogStart,
    file_dialog_end=mod.StandardPixmap.SP_FileDialogEnd,
    file_dialog_to_parent=mod.StandardPixmap.SP_FileDialogToParent,
    file_dialog_new_folder=mod.StandardPixmap.SP_FileDialogNewFolder,
    file_dialog_detailed_view=mod.StandardPixmap.SP_FileDialogDetailedView,
    file_dialog_info_view=mod.StandardPixmap.SP_FileDialogInfoView,
    file_dialog_contents_view=mod.StandardPixmap.SP_FileDialogContentsView,
    file_dialog_list_view=mod.StandardPixmap.SP_FileDialogListView,
    file_dialog_back=mod.StandardPixmap.SP_FileDialogBack,
    dockwidget_close_button=mod.StandardPixmap.SP_DockWidgetCloseButton,
    toolbar_horizontal_extension=mod.StandardPixmap.SP_ToolBarHorizontalExtensionButton,
    toolbar_vertical_extension=mod.StandardPixmap.SP_ToolBarVerticalExtensionButton,
    dialog_ok=mod.StandardPixmap.SP_DialogOkButton,
    dialog_cancel=mod.StandardPixmap.SP_DialogCancelButton,
    dialog_help=mod.StandardPixmap.SP_DialogHelpButton,
    dialog_open=mod.StandardPixmap.SP_DialogOpenButton,
    dialog_save=mod.StandardPixmap.SP_DialogSaveButton,
    dialog_close=mod.StandardPixmap.SP_DialogCloseButton,
    dialog_apply=mod.StandardPixmap.SP_DialogApplyButton,
    dialor_reset=mod.StandardPixmap.SP_DialogResetButton,
    dialog_discard=mod.StandardPixmap.SP_DialogDiscardButton,
    dialog_yes=mod.StandardPixmap.SP_DialogYesButton,
    dialog_no=mod.StandardPixmap.SP_DialogNoButton,
    arrow_up=mod.StandardPixmap.SP_ArrowUp,
    arrow_down=mod.StandardPixmap.SP_ArrowDown,
    arrow_left=mod.StandardPixmap.SP_ArrowLeft,
    arrow_right=mod.StandardPixmap.SP_ArrowRight,
    arrow_back=mod.StandardPixmap.SP_ArrowBack,
    arrow_forward=mod.StandardPixmap.SP_ArrowForward,
    command_link=mod.StandardPixmap.SP_CommandLink,
    vista_shield=mod.StandardPixmap.SP_VistaShield,
    browser_reload=mod.StandardPixmap.SP_BrowserReload,
    browser_stop=mod.StandardPixmap.SP_BrowserStop,
    media_play=mod.StandardPixmap.SP_MediaPlay,
    media_stop=mod.StandardPixmap.SP_MediaStop,
    media_pause=mod.StandardPixmap.SP_MediaPause,
    media_skip_forward=mod.StandardPixmap.SP_MediaSkipForward,
    media_skip_backward=mod.StandardPixmap.SP_MediaSkipBackward,
    media_seek_forward=mod.StandardPixmap.SP_MediaSeekForward,
    media_seek_backward=mod.StandardPixmap.SP_MediaSeekBackward,
    media_volume=mod.StandardPixmap.SP_MediaVolume,
    media_volume_muted=mod.StandardPixmap.SP_MediaVolumeMuted,
    lineedit_clear=mod.StandardPixmap.SP_LineEditClearButton,
    custom_base=mod.StandardPixmap.SP_CustomBase,
    dialog_yes_to_all=mod.StandardPixmap.SP_DialogYesToAllButton,
    dialog_no_to_all=mod.StandardPixmap.SP_DialogNoToAllButton,
    dialog_save_all=mod.StandardPixmap.SP_DialogSaveAllButton,
    dialog_abort=mod.StandardPixmap.SP_DialogAbortButton,
    dialog_retry=mod.StandardPixmap.SP_DialogRetryButton,
    dialog_ignore=mod.StandardPixmap.SP_DialogIgnoreButton,
    restore_defaults=mod.StandardPixmap.SP_RestoreDefaultsButton,
    tab_close=mod.StandardPixmap.SP_TabCloseButton,
)

ComplexControlStr = Literal[
    "spinbox",
    "combobox",
    "scrollbar",
    "slider",
    "toolbutton",
    "titlebar",
    "groupbox",
    "dial",
    "mdi_controls",
    "custom_base",
]


COMPLEX_CONTROL: bidict[ComplexControlStr, mod.ComplexControl] = bidict(
    spinbox=mod.ComplexControl.CC_SpinBox,
    combobox=mod.ComplexControl.CC_ComboBox,
    scrollbar=mod.ComplexControl.CC_ScrollBar,
    slider=mod.ComplexControl.CC_Slider,
    toolbutton=mod.ComplexControl.CC_ToolButton,
    titlebar=mod.ComplexControl.CC_TitleBar,
    groupbox=mod.ComplexControl.CC_GroupBox,
    dial=mod.ComplexControl.CC_Dial,
    mdi_controls=mod.ComplexControl.CC_MdiControls,
    custom_base=mod.ComplexControl.CC_CustomBase,
)

ControlElementStr = Literal[
    "push_button",
    "push_button_bevel",
    "push_button_label",
    "dock_widget_title",
    "splitter",
    "check_box",
    "check_box_label",
    "radio_button",
    "radio_button_label",
    "tab_bar_tab",
    "tab_bar_tab_shape",
    "tab_bar_tab_label",
    "progress_bar",
    "progress_bar_groove",
    "progress_bar_contents",
    "progress_bar_label",
    "tool_button_label",
    "menu_bar_item",
    "menu_bar_empty_area",
    "menu_item",
    "menu_scroller",
    "menu_tear_off",
    "menu_empty_area",
    "menu_h_margin",
    "menu_v_margin",
    "tool_box_tab",
    "size_grip",
    "header",
    "header_section",
    "header_label",
    "scroll_bar_add_line",
    "scroll_bar_sub_line",
    "scroll_bar_add_page",
    "scroll_bar_sub_page",
    "scroll_bar_slider",
    "scroll_bar_first",
    "scroll_bar_last",
    "rubber_band",
    "focus_frame",
    "item_view_item",
    "custom_base",
    "combo_box_label",
    "tool_bar",
    "tool_box_tab_shape",
    "tool_box_tab_label",
    "header_empty_area",
    "shaped_frame",
]

CONTROL_ELEMENT: bidict[ControlElementStr, mod.ControlElement] = bidict(
    push_button=mod.ControlElement.CE_PushButton,
    push_button_bevel=mod.ControlElement.CE_PushButtonBevel,
    push_button_label=mod.ControlElement.CE_PushButtonLabel,
    dock_widget_title=mod.ControlElement.CE_DockWidgetTitle,
    splitter=mod.ControlElement.CE_Splitter,
    check_box=mod.ControlElement.CE_CheckBox,
    check_box_label=mod.ControlElement.CE_CheckBoxLabel,
    radio_button=mod.ControlElement.CE_RadioButton,
    radio_button_label=mod.ControlElement.CE_RadioButtonLabel,
    tab_bar_tab=mod.ControlElement.CE_TabBarTab,
    tab_bar_tab_shape=mod.ControlElement.CE_TabBarTabShape,
    tab_bar_tab_label=mod.ControlElement.CE_TabBarTabLabel,
    progress_bar=mod.ControlElement.CE_ProgressBar,
    progress_bar_groove=mod.ControlElement.CE_ProgressBarGroove,
    progress_bar_contents=mod.ControlElement.CE_ProgressBarContents,
    progress_bar_label=mod.ControlElement.CE_ProgressBarLabel,
    tool_button_label=mod.ControlElement.CE_ToolButtonLabel,
    menu_bar_item=mod.ControlElement.CE_MenuBarItem,
    menu_bar_empty_area=mod.ControlElement.CE_MenuBarEmptyArea,
    menu_item=mod.ControlElement.CE_MenuItem,
    menu_scroller=mod.ControlElement.CE_MenuScroller,
    menu_tear_off=mod.ControlElement.CE_MenuTearoff,
    menu_empty_area=mod.ControlElement.CE_MenuEmptyArea,
    menu_h_margin=mod.ControlElement.CE_MenuHMargin,
    menu_v_margin=mod.ControlElement.CE_MenuVMargin,
    tool_box_tab=mod.ControlElement.CE_ToolBoxTab,
    size_grip=mod.ControlElement.CE_SizeGrip,
    header=mod.ControlElement.CE_Header,
    header_section=mod.ControlElement.CE_HeaderSection,
    header_label=mod.ControlElement.CE_HeaderLabel,
    scroll_bar_add_line=mod.ControlElement.CE_ScrollBarAddLine,
    scroll_bar_sub_line=mod.ControlElement.CE_ScrollBarSubLine,
    scroll_bar_add_page=mod.ControlElement.CE_ScrollBarAddPage,
    scroll_bar_sub_page=mod.ControlElement.CE_ScrollBarSubPage,
    scroll_bar_slider=mod.ControlElement.CE_ScrollBarSlider,
    scroll_bar_first=mod.ControlElement.CE_ScrollBarFirst,
    scroll_bar_last=mod.ControlElement.CE_ScrollBarLast,
    rubber_band=mod.ControlElement.CE_RubberBand,
    focus_frame=mod.ControlElement.CE_FocusFrame,
    item_view_item=mod.ControlElement.CE_ItemViewItem,
    custom_base=mod.ControlElement.CE_CustomBase,
    combo_box_label=mod.ControlElement.CE_ComboBoxLabel,
    tool_bar=mod.ControlElement.CE_ToolBar,
    tool_box_tab_shape=mod.ControlElement.CE_ToolBoxTabShape,
    tool_box_tab_label=mod.ControlElement.CE_ToolBoxTabLabel,
    header_empty_area=mod.ControlElement.CE_HeaderEmptyArea,
    shaped_frame=mod.ControlElement.CE_ShapedFrame,
)

SubControlStr = Literal[
    "none",
    "scrollbar_add_line",
    "scrollbar_sub_line",
    "scrollbar_add_page",
    "scrollbar_sub_page",
    "scrollbar_first",
    "scrollbar_last",
    "scrollbar_slider",
    "scrollbar_groove",
    "spinbox_up",
    "spinbox_down",
    "spinbox_frame",
    "spinbox_edit_field",
    "combobox_edit_field",
    "combobox_arrow",
    "combobox_frame",
    "combobox_list_box_popup",
    "slider_groove",
    "slider_handle",
    "slider_tickmarks",
    "toolbutton",
    "toolbutton_menu",
    "titlebar_sys_menu",
    "titlebar_min_button",
    "titlebar_max_button",
    "titlebar_close_button",
    "titlebar_label",
    "titlebar_normal_button",
    "titlebar_shade_button",
    "titlebar_unshade_button",
    "titlebar_context_help_button",
    "dial_handle",
    "dial_groove",
    "dial_tickmarks",
    "groupbox_frame",
    "groupbox_label",
    "groupbox_checkbox",
    "groupbox_contents",
    "mdi_normal_button",
    "mdi_min_button",
    "mdi_close_button",
    "all",
]

SUB_CONTROL: dict[SubControlStr, mod.SubControl] = dict(
    none=mod.SubControl.SC_None,
    scrollbar_add_line=mod.SubControl.SC_ScrollBarAddLine,
    scrollbar_sub_line=mod.SubControl.SC_ScrollBarSubLine,
    scrollbar_add_page=mod.SubControl.SC_ScrollBarAddPage,
    scrollbar_sub_page=mod.SubControl.SC_ScrollBarSubPage,
    scrollbar_first=mod.SubControl.SC_ScrollBarFirst,
    scrollbar_last=mod.SubControl.SC_ScrollBarLast,
    scrollbar_slider=mod.SubControl.SC_ScrollBarSlider,
    scrollbar_groove=mod.SubControl.SC_ScrollBarGroove,
    spinbox_up=mod.SubControl.SC_SpinBoxUp,
    spinbox_down=mod.SubControl.SC_SpinBoxDown,
    spinbox_frame=mod.SubControl.SC_SpinBoxFrame,
    spinbox_edit_field=mod.SubControl.SC_SpinBoxEditField,
    combobox_edit_field=mod.SubControl.SC_ComboBoxEditField,
    combobox_arrow=mod.SubControl.SC_ComboBoxArrow,
    combobox_frame=mod.SubControl.SC_ComboBoxFrame,
    combobox_list_box_popup=mod.SubControl.SC_ComboBoxListBoxPopup,
    slider_groove=mod.SubControl.SC_SliderGroove,
    slider_handle=mod.SubControl.SC_SliderHandle,
    slider_tickmarks=mod.SubControl.SC_SliderTickmarks,
    toolbutton=mod.SubControl.SC_ToolButton,
    toolbutton_menu=mod.SubControl.SC_ToolButtonMenu,
    titlebar_sys_menu=mod.SubControl.SC_TitleBarSysMenu,
    titlebar_min_button=mod.SubControl.SC_TitleBarMinButton,
    titlebar_max_button=mod.SubControl.SC_TitleBarMaxButton,
    titlebar_close_button=mod.SubControl.SC_TitleBarCloseButton,
    titlebar_label=mod.SubControl.SC_TitleBarLabel,
    titlebar_normal_button=mod.SubControl.SC_TitleBarNormalButton,
    titlebar_shade_button=mod.SubControl.SC_TitleBarShadeButton,
    titlebar_unshade_button=mod.SubControl.SC_TitleBarUnshadeButton,
    titlebar_context_help_button=mod.SubControl.SC_TitleBarContextHelpButton,
    dial_handle=mod.SubControl.SC_DialHandle,
    dial_groove=mod.SubControl.SC_DialGroove,
    dial_tickmarks=mod.SubControl.SC_DialTickmarks,
    groupbox_frame=mod.SubControl.SC_GroupBoxFrame,
    groupbox_label=mod.SubControl.SC_GroupBoxLabel,
    groupbox_checkbox=mod.SubControl.SC_GroupBoxCheckBox,
    groupbox_contents=mod.SubControl.SC_GroupBoxContents,
    mdi_normal_button=mod.SubControl.SC_MdiNormalButton,
    mdi_min_button=mod.SubControl.SC_MdiMinButton,
    mdi_close_button=mod.SubControl.SC_MdiCloseButton,
    all=mod.SubControl.SC_All,
)

PrimitiveElementStr = Literal[
    "panel_button_command",
    "frame_default_button",
    "panel_button_bevel",
    "panel_button_tool",
    "panel_line_edit",
    "indicator_button_drop_down",
    "frame_focus_rect",
    "indicator_arrow_up",
    "indicator_arrow_down",
    "indicator_arrow_right",
    "indicator_arrow_left",
    "indicator_spin_up",
    "indicator_spin_down",
    "indicator_spin_plus",
    "indicator_spin_minus",
    "indicator_item_view_item_check",
    "indicator_check_box",
    "indicator_radio_button",
    "indicator_dock_widget_resize_handle",
    "frame",
    "frame_menu",
    "panel_menu_bar",
    "panel_scroll_area_corner",
    "frame_dock_widget",
    "frame_tab_widget",
    "frame_line_edit",
    "frame_group_box",
    "frame_button_bevel",
    "frame_button_tool",
    "indicator_header_arrow",
    "frame_status_bar_item",
    "frame_window",
    "indicator_menu_check_mark",
    "indicator_progress_chunk",
    "indicator_branch",
    "indicator_tool_bar_handle",
    "indicator_tool_bar_separator",
    "panel_tool_bar",
    "panel_tip_label",
    "frame_tab_bar_base",
    "indicator_tab_tear_left",
    "indicator_tab_tear_right",
    "indicator_column_view_arrow",
    "widget",
    "custom_base",
    "indicator_item_view_item_drop",
    "panel_item_view_item",
    "panel_item_view_row",
    "panel_status_bar",
    "indicator_tab_close",
    "panel_menu",
]

pe = mod.PrimitiveElement

PRIMITIVE_ELEMENT: bidict[PrimitiveElementStr, pe] = bidict(
    panel_button_command=pe.PE_PanelButtonCommand,
    frame_default_button=pe.PE_FrameDefaultButton,
    panel_button_bevel=pe.PE_PanelButtonBevel,
    panel_button_tool=pe.PE_PanelButtonTool,
    panel_line_edit=pe.PE_PanelLineEdit,
    indicator_button_drop_down=pe.PE_IndicatorButtonDropDown,
    frame_focus_rect=pe.PE_FrameFocusRect,
    indicator_arrow_up=pe.PE_IndicatorArrowUp,
    indicator_arrow_down=pe.PE_IndicatorArrowDown,
    indicator_arrow_right=pe.PE_IndicatorArrowRight,
    indicator_arrow_left=pe.PE_IndicatorArrowLeft,
    indicator_spin_up=pe.PE_IndicatorSpinUp,
    indicator_spin_down=pe.PE_IndicatorSpinDown,
    indicator_spin_plus=pe.PE_IndicatorSpinPlus,
    indicator_spin_minus=pe.PE_IndicatorSpinMinus,
    indicator_item_view_item_check=pe.PE_IndicatorItemViewItemCheck,
    indicator_check_box=pe.PE_IndicatorCheckBox,
    indicator_radio_button=pe.PE_IndicatorRadioButton,
    indicator_dock_widget_resize_handle=pe.PE_IndicatorDockWidgetResizeHandle,
    frame=pe.PE_Frame,
    frame_menu=pe.PE_FrameMenu,
    panel_menu_bar=pe.PE_PanelMenuBar,
    panel_scroll_area_corner=pe.PE_PanelScrollAreaCorner,
    frame_dock_widget=pe.PE_FrameDockWidget,
    frame_tab_widget=pe.PE_FrameTabWidget,
    frame_line_edit=pe.PE_FrameLineEdit,
    frame_group_box=pe.PE_FrameGroupBox,
    frame_button_bevel=pe.PE_FrameButtonBevel,
    frame_button_tool=pe.PE_FrameButtonTool,
    indicator_header_arrow=pe.PE_IndicatorHeaderArrow,
    frame_status_bar_item=pe.PE_FrameStatusBarItem,
    frame_window=pe.PE_FrameWindow,
    indicator_menu_check_mark=pe.PE_IndicatorMenuCheckMark,
    indicator_progress_chunk=pe.PE_IndicatorProgressChunk,
    indicator_branch=pe.PE_IndicatorBranch,
    indicator_tool_bar_handle=pe.PE_IndicatorToolBarHandle,
    indicator_tool_bar_separator=pe.PE_IndicatorToolBarSeparator,
    panel_tool_bar=pe.PE_PanelToolBar,
    panel_tip_label=pe.PE_PanelTipLabel,
    frame_tab_bar_base=pe.PE_FrameTabBarBase,
    indicator_tab_tear_left=pe.PE_IndicatorTabTearLeft,
    indicator_tab_tear_right=pe.PE_IndicatorTabTearRight,
    indicator_column_view_arrow=pe.PE_IndicatorColumnViewArrow,
    widget=pe.PE_Widget,
    custom_base=pe.PE_CustomBase,
    indicator_item_view_item_drop=pe.PE_IndicatorItemViewItemDrop,
    panel_item_view_item=pe.PE_PanelItemViewItem,
    panel_item_view_row=pe.PE_PanelItemViewRow,
    panel_status_bar=pe.PE_PanelStatusBar,
    indicator_tab_close=pe.PE_IndicatorTabClose,
    panel_menu=pe.PE_PanelMenu,
)

PixelMetricStr = Literal[
    "button_margin",
    "dock_widget_title_bar_button_margin",
    "button_default_indicator",
    "menu_button_indicator",
    "button_shift_horizontal",
    "button_shift_vertical",
    "default_frame_width",
    "spinbox_frame_width",
    "combobox_frame_width",
    "mdi_subwindow_frame_width",
    "mdi_subwindow_minimized_width",
    "layout_left_margin",
    "layout_top_margin",
    "layout_right_margin",
    "layout_bottom_margin",
    "layout_horizontal_spacing",
    "layout_vertical_spacing",
    "maximum_drag_distance",
    "scrollbar_extent",
    "scrollbar_slider_min",
    "slider_thickness",
    "slider_control_thickness",
    "slider_length",
    "slider_tickmark_offset",
    "slider_space_available",
    "dock_widget_separator_extent",
    "dock_widget_handle_extent",
    "dock_widget_frame_width",
    "dock_widget_title_margin",
    "menu_bar_panel_width",
    "menu_bar_item_spacing",
    "menu_bar_h_margin",
    "menu_bar_v_margin",
    "tool_bar_frame_width",
    "tool_bar_handle_extent",
    "tool_bar_item_margin",
    "tool_bar_item_spacing",
    "tool_bar_separator_extent",
    "tool_bar_extension_extent",
    "tab_bar_tab_overlap",
    "tab_bar_tab_h_space",
    "tab_bar_tab_v_space",
    "tab_bar_base_height",
    "tab_bar_base_overlap",
    "tab_bar_scroll_button_width",
    "tab_bar_tab_shift_horizontal",
    "tab_bar_tab_shift_vertical",
    "progress_bar_chunk_width",
    "splitter_width",
    "title_bar_height",
    "indicator_width",
    "indicator_height",
    "exclusive_indicator_width",
    "exclusive_indicator_height",
    "menu_panel_width",
    "menu_h_margin",
    "menu_v_margin",
    "menu_scroller_height",
    "menu_tearoff_height",
    "menu_desktop_frame_width",
    "header_mark_size",
    "header_grip_margin",
    "header_margin",
    "spin_box_slider_height",
    "tool_bar_icon_size",
    "small_icon_size",
    "large_icon_size",
    "focus_frame_h_margin",
    "focus_frame_v_margin",
    "icon_view_icon_size",
    "list_view_icon_size",
    "tool_tip_label_frame_width",
    "check_box_label_spacing",
    "radio_button_label_spacing",
    "tab_bar_icon_size",
    "size_grip_size",
    "message_box_icon_size",
    "button_icon_size",
    "text_cursor_width",
    "tabbar_scrollbuttonoverlap",
    "tab_close_indicator_width",
    "tab_close_indicator_height",
    "scroll_view_scroll_bar_spacing",
    "scroll_view_scroll_bar_overlap",
    "sub_menu_overlap",
    "tree_view_indentation",
    "header_default_section_size_horizontal",
    "header_default_section_size_vertical",
    "title_bar_button_icon_size",
    "title_bar_button_size",
    "line_edit_icon_size",
    "line_edit_icon_margin",
]

pm = mod.PixelMetric

PIXEL_METRIC: bidict[PixelMetricStr, pm] = bidict(
    button_margin=pm.PM_ButtonMargin,
    dock_widget_title_bar_button_margin=pm.PM_DockWidgetTitleBarButtonMargin,
    button_default_indicator=pm.PM_ButtonDefaultIndicator,
    menu_button_indicator=pm.PM_MenuButtonIndicator,
    button_shift_horizontal=pm.PM_ButtonShiftHorizontal,
    button_shift_vertical=pm.PM_ButtonShiftVertical,
    default_frame_width=pm.PM_DefaultFrameWidth,
    spinbox_frame_width=pm.PM_SpinBoxFrameWidth,
    combobox_frame_width=pm.PM_ComboBoxFrameWidth,
    mdi_subwindow_frame_width=pm.PM_MdiSubWindowFrameWidth,
    mdi_subwindow_minimized_width=pm.PM_MdiSubWindowMinimizedWidth,
    layout_left_margin=pm.PM_LayoutLeftMargin,
    layout_top_margin=pm.PM_LayoutTopMargin,
    layout_right_margin=pm.PM_LayoutRightMargin,
    layout_bottom_margin=pm.PM_LayoutBottomMargin,
    layout_horizontal_spacing=pm.PM_LayoutHorizontalSpacing,
    layout_vertical_spacing=pm.PM_LayoutVerticalSpacing,
    maximum_drag_distance=pm.PM_MaximumDragDistance,
    scrollbar_extent=pm.PM_ScrollBarExtent,
    scrollbar_slider_min=pm.PM_ScrollBarSliderMin,
    slider_thickness=pm.PM_SliderThickness,
    slider_control_thickness=pm.PM_SliderControlThickness,
    slider_length=pm.PM_SliderLength,
    slider_tickmark_offset=pm.PM_SliderTickmarkOffset,
    slider_space_available=pm.PM_SliderSpaceAvailable,
    dock_widget_separator_extent=pm.PM_DockWidgetSeparatorExtent,
    dock_widget_handle_extent=pm.PM_DockWidgetHandleExtent,
    dock_widget_frame_width=pm.PM_DockWidgetFrameWidth,
    dock_widget_title_margin=pm.PM_DockWidgetTitleMargin,
    menu_bar_panel_width=pm.PM_MenuBarPanelWidth,
    menu_bar_item_spacing=pm.PM_MenuBarItemSpacing,
    menu_bar_h_margin=pm.PM_MenuBarHMargin,
    menu_bar_v_margin=pm.PM_MenuBarVMargin,
    tool_bar_frame_width=pm.PM_ToolBarFrameWidth,
    tool_bar_handle_extent=pm.PM_ToolBarHandleExtent,
    tool_bar_item_margin=pm.PM_ToolBarItemMargin,
    tool_bar_item_spacing=pm.PM_ToolBarItemSpacing,
    tool_bar_separator_extent=pm.PM_ToolBarSeparatorExtent,
    tool_bar_extension_extent=pm.PM_ToolBarExtensionExtent,
    tab_bar_tab_overlap=pm.PM_TabBarTabOverlap,
    tab_bar_tab_h_space=pm.PM_TabBarTabHSpace,
    tab_bar_tab_v_space=pm.PM_TabBarTabVSpace,
    tab_bar_base_height=pm.PM_TabBarBaseHeight,
    tab_bar_base_overlap=pm.PM_TabBarBaseOverlap,
    tab_bar_scroll_button_width=pm.PM_TabBarScrollButtonWidth,
    tab_bar_tab_shift_horizontal=pm.PM_TabBarTabShiftHorizontal,
    tab_bar_tab_shift_vertical=pm.PM_TabBarTabShiftVertical,
    progress_bar_chunk_width=pm.PM_ProgressBarChunkWidth,
    splitter_width=pm.PM_SplitterWidth,
    title_bar_height=pm.PM_TitleBarHeight,
    indicator_width=pm.PM_IndicatorWidth,
    indicator_height=pm.PM_IndicatorHeight,
    exclusive_indicator_width=pm.PM_ExclusiveIndicatorWidth,
    exclusive_indicator_height=pm.PM_ExclusiveIndicatorHeight,
    menu_panel_width=pm.PM_MenuPanelWidth,
    menu_h_margin=pm.PM_MenuHMargin,
    menu_v_margin=pm.PM_MenuVMargin,
    menu_scroller_height=pm.PM_MenuScrollerHeight,
    menu_tearoff_height=pm.PM_MenuTearoffHeight,
    menu_desktop_frame_width=pm.PM_MenuDesktopFrameWidth,
    header_mark_size=pm.PM_HeaderMarkSize,
    header_grip_margin=pm.PM_HeaderGripMargin,
    header_margin=pm.PM_HeaderMargin,
    spin_box_slider_height=pm.PM_SpinBoxSliderHeight,
    tool_bar_icon_size=pm.PM_ToolBarIconSize,
    small_icon_size=pm.PM_SmallIconSize,
    large_icon_size=pm.PM_LargeIconSize,
    focus_frame_h_margin=pm.PM_FocusFrameHMargin,
    focus_frame_v_margin=pm.PM_FocusFrameVMargin,
    icon_view_icon_size=pm.PM_IconViewIconSize,
    list_view_icon_size=pm.PM_ListViewIconSize,
    tool_tip_label_frame_width=pm.PM_ToolTipLabelFrameWidth,
    check_box_label_spacing=pm.PM_CheckBoxLabelSpacing,
    radio_button_label_spacing=pm.PM_RadioButtonLabelSpacing,
    tab_bar_icon_size=pm.PM_TabBarIconSize,
    size_grip_size=pm.PM_SizeGripSize,
    message_box_icon_size=pm.PM_MessageBoxIconSize,
    button_icon_size=pm.PM_ButtonIconSize,
    text_cursor_width=pm.PM_TextCursorWidth,
    tabbar_scrollbuttonoverlap=pm.PM_TabBar_ScrollButtonOverlap,
    tab_close_indicator_width=pm.PM_TabCloseIndicatorWidth,
    tab_close_indicator_height=pm.PM_TabCloseIndicatorHeight,
    scroll_view_scroll_bar_spacing=pm.PM_ScrollView_ScrollBarSpacing,
    scroll_view_scroll_bar_overlap=pm.PM_ScrollView_ScrollBarOverlap,
    sub_menu_overlap=pm.PM_SubMenuOverlap,
    tree_view_indentation=pm.PM_TreeViewIndentation,
    header_default_section_size_horizontal=pm.PM_HeaderDefaultSectionSizeHorizontal,
    header_default_section_size_vertical=pm.PM_HeaderDefaultSectionSizeVertical,
    title_bar_button_icon_size=pm.PM_TitleBarButtonIconSize,
    title_bar_button_size=pm.PM_TitleBarButtonSize,
    line_edit_icon_size=pm.PM_LineEditIconSize,
    line_edit_icon_margin=pm.PM_LineEditIconMargin,
)

StateStr = Literal[
    "none",
    "active",
    "auto_raise",
    "children",
    "down_arrow",
    "editing",
    "enabled",
    "has_edit_focus",
    "has_focus",
    "horizontal",
    "keyboard_focus_change",
    "mouse_over",
    "no_change",
    "off",
    "on",
    "raised",
    "read_only",
    "selected",
    "item",
    "open",
    "sibling",
    "sunken",
    "up_arrow",
    "mini",
    "small",
]

st = mod.StateFlag

State: bidict[StateStr, st] = bidict(
    none=st.State_None,
    active=st.State_Active,
    auto_raise=st.State_AutoRaise,
    children=st.State_Children,
    down_arrow=st.State_DownArrow,
    editing=st.State_Editing,
    enabled=st.State_Enabled,
    # has_edit_focus=st.State_HasEditFocus,
    has_focus=st.State_HasFocus,
    horizontal=st.State_Horizontal,
    keyboard_focus_change=st.State_KeyboardFocusChange,
    mouse_over=st.State_MouseOver,
    no_change=st.State_NoChange,
    off=st.State_Off,
    on=st.State_On,
    raised=st.State_Raised,
    read_only=st.State_ReadOnly,
    selected=st.State_Selected,
    item=st.State_Item,
    open=st.State_Open,
    sibling=st.State_Sibling,
    sunken=st.State_Sunken,
    up_arrow=st.State_UpArrow,
    mini=st.State_Mini,
    small=st.State_Small,
)


class StyleMixin(core.ObjectMixin):
    def draw_primitive(
        self,
        element: SubControlStr | mod.SubControl,
        option: widgets.QStyleOption,
        painter: QtGui.QPainter,
        widget: widgets.QWidget | None = None,
    ):
        self.drawPrimitive(
            PRIMITIVE_ELEMENT.get_enum_value(element), option, painter, widget
        )

    def draw_control(
        self,
        control: ControlElementStr | mod.ControlElement,
        option: widgets.QStyleOption,
        painter: QtGui.QPainter,
        widget: widgets.QWidget | None = None,
    ):
        self.drawPrimitive(
            CONTROL_ELEMENT.get_enum_value(control), option, painter, widget
        )

    def get_layout_spacing(
        self,
        control_1: widgets.sizepolicy.ControlTypeStr | widgets.QSizePolicy.ControlType,
        control_2: widgets.sizepolicy.ControlTypeStr | widgets.QSizePolicy.ControlType,
        orientation: constants.OrientationStr | constants.Orientation,
        option_or_widget: widgets.QStyleOption | widgets.QWidget | None = None,
    ):
        c1 = widgets.sizepolicy.CONTROL_TYPE.get_enum_value(control_1)
        c2 = widgets.sizepolicy.CONTROL_TYPE.get_enum_value(control_2)
        o = constants.ORIENTATION.get_enum_value(orientation)
        match option_or_widget:
            case widgets.QWidget():
                return self.layoutSpacing(c1, c2, o, None, option_or_widget)
            case widgets.QStyleOption() | None:
                return self.layoutSpacing(c1, c2, o, option_or_widget)
            case _:
                raise ValueError(option_or_widget)


class Style(StyleMixin, widgets.QStyle):
    """Abstract base class that encapsulates the look and feel of a GUI."""



from __future__ import annotations

from prettyqt import core, gui, widgets


class FaderWidget(widgets.Widget):
    pixmap_opacity = 1.0

    def __init__(
        self,
        old_widget: widgets.QWidget,
        new_widget: widgets.QWidget,
        duration: int = 300,
    ):
        super().__init__(new_widget)

        pr = gui.Window().devicePixelRatio()
        self.old_pixmap = gui.Pixmap(new_widget.size() * pr)
        self.old_pixmap.setDevicePixelRatio(pr)
        old_widget.render(self.old_pixmap)

        self.timeline = core.TimeLine(duration=duration, finished=self.close)
        self.timeline.value_changed.connect(self.animate)
        self.timeline.start()

        self.resize(new_widget.size())
        self.show()

    @classmethod
    def setup_example(cls):
        return None

    def paintEvent(self, event):
        with gui.Painter(self) as painter:
            painter.setOpacity(self.pixmap_opacity)
            painter.drawPixmap(0, 0, self.old_pixmap)

    def animate(self, value: float):
        self.pixmap_opacity = 1.0 - value
        self.repaint()


if __name__ == "__main__":
    app = widgets.app()
    w1 = widgets.RadioButton("fdsfsdf")
    w2 = widgets.PlainTextEdit("fbbb")
    widget = FaderWidget(w1, w2)
    with app.debug_mode():
        app.exec()



from __future__ import annotations

from collections.abc import Callable
import contextlib
import functools
import logging
import time

from prettyqt import qt
from prettyqt.qt import QtCore

from .errormessagebox import ErrorMessageBox
from .messagehandler import MessageHandler
from .qobjectdetailsdialog import QObjectDetailsDialog
from .stalker import Stalker
from .tracebackdialog import TracebackDialog
from .proxycomparerwidget import ProxyComparerWidget


logger = logging.getLogger(__name__)


def timeit(func: Callable) -> Callable:
    @functools.wraps(func)
    def timeit_wrapper(*args, **kwargs):
        # logger.info(f'Function {func.__name__}{args} {kwargs}')
        start_time = time.perf_counter()
        result = func(*args, **kwargs)
        end_time = time.perf_counter()
        total_time = end_time - start_time
        logger.info(
            f"Function {func.__name__}{args} {kwargs} Took {total_time:.4f} seconds"
        )
        return result

    return timeit_wrapper


@contextlib.contextmanager
def context_timer():
    now = time.perf_counter()
    yield
    logger.info(time.perf_counter() - now)


def for_all_methods(decorator: Callable) -> Callable:
    def decorate(cls):
        for attr in cls.__dict__:  # there's propably a better way to do this
            if callable(getattr(cls, attr)):
                logger.info(f"decorated {attr}")
                setattr(cls, attr, decorator(getattr(cls, attr)))
        return cls

    return decorate


class QtLogger(logging.Handler):
    def emit(self, record: logging.LogRecord):
        match record.level:
            case logging.DEBUG:
                QtCore.qDebug(self.format(record))
            case logging.INFO:
                QtCore.qInfo(self.format(record))
            case logging.WARNING:
                QtCore.qWarning(self.format(record))
            case logging.CRITICAL:
                QtCore.qCritical(self.format(record))
            case logging.CRITICAL:
                QtCore.qFatal(self.format(record))


def is_deleted(obj) -> bool:
    match qt.API:
        case "pyside6":
            import shiboken6

            return not shiboken6.isValid(obj)
        case "pyqt6":
            from PyQt6 import sip

            return sip.isdeleted(obj)


def example_tree(flatten: bool = False):
    from prettyqt import widgets
    from prettyqt.itemmodels import JsonModel

    dist = [
        dict(
            a=2,
            b={
                "a": 4,
                "b": [1, 2, 3],
                "jkjkjk": "tekjk",
                "sggg": "tekjk",
                "fdfdf": "tekjk",
                "xxxx": "xxx",
            },
        ),
        6,
        "jkjk",
    ]
    table = widgets.TreeView()
    model = JsonModel(dist, parent=table)
    table.setRootIsDecorated(True)
    table.set_model(model)
    if flatten:
        table.proxifier.flatten()
    return table


def example_table(flatten: bool = False):
    from prettyqt import gui, widgets

    data = dict(
        a=["abcdedf", "abcdedf", "abcdedf", "abcdedf", "abcdedf", "abcdedfaa"],
        b=[10000000, 2, 3, 4, 5, 6],
        c=[1, 2, 3, 4, 5, 6],
        d=[100000000, 2, 3, 4, 5, 6],
        e=[1000000, 2, 3, 4, 5, 6],
        f=[1000000, 2, 3, 4, 5, 6],
        g=[1, 2, 3, 4, 5, 6],
        h=[1, 2, 3, 4, 5, 6],
        i=[100000000000000, 2, 3, 4, 5, 6],
        j=[1, 2, 3, 4, 5, 6],
        k=[1, 2, 3, 4, 5, 6],
        jkfsj=[10, 20, 30, 40, 50, 60],
    )
    model = gui.StandardItemModel.from_dict(data)
    table = widgets.TableView()
    table.set_model(model)
    return table


def example_widget():
    from prettyqt import widgets

    widget = widgets.Widget()
    layout = widget.set_layout("horizontal", object_name="main")
    with layout.get_sub_layout("splitter", orientation="horizontal") as layout:
        with layout.get_sub_layout("grid") as layout:
            layout[0, 0] = widgets.PushButton("Grid topleft")
            layout[0, 1] = widgets.RadioButton("Grid topright")
            layout[1, 0:1] = widgets.PushButton("Grid bottom")

        with layout.get_sub_layout("flow", object_name="flow") as layout:
            # with layout.get_sub_layout("frame", title="test"):
            layout += widgets.PushButton("Flow 1")
            layout += widgets.RadioButton("Flow 2")
            layout += widgets.PushButton("Flow 3")
            layout += widgets.RadioButton("Flow 4")
        layout += widgets.PlainTextEdit("Splitter middle")
        layout += widgets.PlainTextEdit("Splitter right")
        with layout.get_sub_layout("splitter", orientation="vertical") as layout:
            # with layout.get_sub_layout("frame", title="test"):
            layout += widgets.PlainTextEdit("Splitter top")
            layout += widgets.PlainTextEdit("Splitter middle")
            with layout.get_sub_layout("scroll", orientation="vertical") as layout:
                layout += widgets.PlainTextEdit("ScrollArea top")
                layout += widgets.PlainTextEdit("ScrollArea middle")
                button = layout.add(widgets.PushButton("ScrollArea Bottom"))
    with layout.get_sub_layout("horizontal", object_name="h1") as layout:
        layout += widgets.PlainTextEdit("HorizontalLayout left")
        layout += widgets.PlainTextEdit("HorizontalLayout right")
        with layout.get_sub_layout("grid") as layout2:
            layout2[0, 0] = widgets.PlainTextEdit("VerticalLayout left")
            layout2[0, 1] = widgets.PlainTextEdit("VerticalLayout right")
            layout2[1, 0:1] = example_table()
    button.clicked.connect(lambda: widget.layout().addWidget(widgets.Label("test")))
    return widget


__all__ = [
    "Stalker",
    "TracebackDialog",
    "ErrorMessageBox",
    "MessageHandler",
    "QObjectDetailsDialog",
    "ProxyComparerWidget",
]



from __future__ import annotations

from typing import Literal

from prettyqt import statemachine
from prettyqt.qt import QtCore
from prettyqt.utils import bidict


sm = statemachine.QStateMachine.Error

ErrorStr = Literal[
    "none",
    "no_initial_state",
    "no_default_state_in_history_state",
    "no_common_ancestor_for_transistion",
    "state_machine_child_Mode_set_to_parallel",
]

ERROR: bidict[ErrorStr, sm] = bidict(
    none=sm.NoError,
    no_initial_state=sm.NoInitialStateError,
    no_default_state_in_history_state=sm.NoDefaultStateInHistoryStateError,
    no_common_ancestor_for_transistion=sm.NoCommonAncestorForTransitionError,
    state_machine_child_Mode_set_to_parallel=sm.StateMachineChildModeSetToParallelError,
)

PriorityStr = Literal["normal", "high"]

PRIORITY: bidict[PriorityStr, statemachine.QStateMachine.EventPriority] = bidict(
    normal=statemachine.QStateMachine.EventPriority.NormalPriority,
    high=statemachine.QStateMachine.EventPriority.HighPriority,
)


class StateMachine(statemachine.state.StateMixin, statemachine.QStateMachine):
    def __add__(self, other: statemachine.QAbstractState) -> StateMachine:
        self.addState(other)
        return self

    def get_error(self) -> ErrorStr:
        return ERROR.inverse[self.error()]

    def post_event(
        self,
        event: QtCore.QEvent,
        priority: PriorityStr | statemachine.QStateMachine.EventPriority = "normal",
    ):
        self.postEvent(event, PRIORITY.get_enum_value(priority))

    def set_global_restore_policy(
        self,
        policy: statemachine.state.RestorePolicyStr | statemachine.State.RestorePolicy,
    ):
        """Set restore policy to use.

        Args:
            policy: restore policy to use
        """
        self.setGlobalRestorePolicy(
            statemachine.state.RESTORE_POLICY.get_enum_value(policy)
        )

    def get_global_restore_policy(self) -> statemachine.state.RestorePolicyStr:
        """Return current restore policy.

        Returns:
            restore policy
        """
        return statemachine.state.RESTORE_POLICY.inverse[self.globalRestorePolicy()]



from __future__ import annotations

import contextlib
import enum
import functools
import inspect
import logging
import types

from typing import get_args

from prettyqt import constants, core, gui, itemmodels


logger = logging.getLogger(__name__)


@functools.cache
def get_comments(klass: type) -> str | None:
    return inspect.getcomments(klass)


@functools.cache
def get_doc(klass: type) -> str | None:
    return inspect.getdoc(klass)


@functools.cache
def get_file(klass: type) -> str | None:
    with contextlib.suppress(TypeError):
        return inspect.getfile(klass)


@functools.cache
def get_source(klass: type) -> str:
    return inspect.getsource(klass)


@functools.cache
def get_module(klass: type):
    return inspect.getmodule(klass).__name__


SOURCE_FONT = gui.Font.mono(as_qt=True)


class BaseClassTreeModel(itemmodels.TreeModel):
    """Base Tree Model to display class tree structures."""

    HEADER = ["Name", "Docstrings", "Module", "Comments", "File", "Is Abstract"]

    @core.Enum
    class Roles(enum.IntEnum):
        """Item roles."""

        SourceRole = constants.USER_ROLE + 1

    def columnCount(self, parent=None):
        return len(self.HEADER)

    def headerData(
        self,
        section: int,
        orientation: constants.Orientation,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ) -> str | None:
        match orientation, role, section:
            case constants.HORIZONTAL, constants.DISPLAY_ROLE, _:
                return self.HEADER[section]
        return None

    def data(self, index: core.ModelIndex, role=constants.DISPLAY_ROLE):
        if not index.isValid():
            return None
        klass = self.data_by_index(index).obj
        match role, index.column():
            case constants.DISPLAY_ROLE, 0:
                return "Union" if isinstance(klass, types.UnionType) else klass.__name__
            case constants.DISPLAY_ROLE, 1:
                return get_doc(klass)
            case constants.DISPLAY_ROLE, 2:
                return get_module(klass)
            case constants.DISPLAY_ROLE, 3:
                return get_comments(klass)
            case constants.DISPLAY_ROLE, 4:
                return get_file(klass)
            case constants.CHECKSTATE_ROLE, 5:
                return self.to_checkstate(inspect.isabstract(klass))
            case constants.USER_ROLE, _:
                return klass
            case constants.FONT_ROLE, 1 | 3:
                return SOURCE_FONT
            case self.Roles.SourceRole, _:
                return get_source(klass)


class SubClassTreeModel(BaseClassTreeModel):
    """Model to display the subclass tree of a python class.

    ```py
    from prettyqt import itemmodels, widgets

    app = widgets.app()
    widget = widgets.TreeView()
    model = itemmodels.SubClassTreeModel(core.AbstractItemModelMixin)
    widget.set_model(model)
    widget.show()
    ```

    <figure markdown>
      ![Image title](abstractitemmodelmixin_subclasses.png)
    </figure>
    """

    SUPPORTS = type | types.UnionType

    @classmethod
    def supports(cls, instance) -> bool:
        return isinstance(instance, type | types.UnionType)

    def _fetch_object_children(
        self, item: SubClassTreeModel.TreeItem
    ) -> list[SubClassTreeModel.TreeItem]:
        if isinstance(item.obj, types.UnionType):
            return [self.TreeItem(obj=i) for i in get_args(item.obj)]
        return [self.TreeItem(obj=i) for i in item.obj.__subclasses__()]

    def _has_children(self, item: SubClassTreeModel.TreeItem) -> bool:
        if item.obj is None:
            return False
        if isinstance(item.obj, types.UnionType):
            return True
        return len(item.obj.__subclasses__()) > 0


class ParentClassTreeModel(BaseClassTreeModel):
    """Tree model to display the parent class tree of a python class.

    === "Parentclass tree"

        ```py
        from prettyqt import itemmodels, widgets

        app = widgets.app()
        widget = widgets.TreeView()
        model = itemmodels.ParentClassTreeModel(widgets.TreeWidget)
        widget.set_model(model)
        widget.show()
        ```

        <figure markdown>
          ![Image title](treewidget_parentclasses.png)
        </figure>

    === "MRO tree"

        ```py
        from prettyqt import itemmodels, widgets

        app = widgets.app()
        widget = widgets.TreeView()
        model = itemmodels.ParentClassTreeModel(widgets.TreeWidget, mro=True)
        widget.set_model(model)
        widget.show()
        ```

        <figure markdown>
          ![Image title](treewidget_mro.png)
        </figure>
    """

    SUPPORTS = type

    def __init__(self, *args, **kwargs):
        self._show_mro = False
        super().__init__(*args, **kwargs)

    @classmethod
    def supports(cls, instance) -> bool:
        return isinstance(instance, type)

    def _fetch_object_children(
        self, item: ParentClassTreeModel.TreeItem
    ) -> list[ParentClassTreeModel.TreeItem]:
        if self._show_mro:
            return [self.TreeItem(obj=i) for i in item.obj.mro()[1:]]
        else:
            return [self.TreeItem(obj=i) for i in item.obj.__bases__]

    def _has_children(self, item: ParentClassTreeModel.TreeItem) -> bool:
        if item.obj is None:
            return False
        return len(item.obj.__bases__) > 0

    def set_show_mro(self, show: bool):
        """Toggles mro mode on or off.

        Arguments:
            show: toggle mro mode on/off
        """
        self._show_mro = show

    def get_show_mro(self) -> bool:
        """Returns True if mro mode is turned on."""
        return self._show_mro

    show_mro = core.Property(
        bool,
        get_show_mro,
        set_show_mro,
        doc="Whether children items should be the ordered mro.",
    )


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    view = widgets.TreeView()
    view.setRootIsDecorated(True)
    model = SubClassTreeModel(core.ObjectMixin | core.Object, show_root=True, parent=view)
    view.set_model(model)
    # view.proxifier[0].modify(
    #     lambda x: gui.QColor("blue"),
    #     role=constants.BACKGROUND_ROLE,
    #     selector=lambda x: issubclass(x, core.QObject),
    #     selector_role=constants.USER_ROLE,
    # )
    view.setEditTriggers(view.EditTrigger.AllEditTriggers)
    view.set_delegate("editor")
    view.resize(1000, 1000)
    view.show()
    with app.debug_mode():
        app.exec()



"""Module containing stuff to help buidling docs.

This module builds on mknodes and contains some new nodes to help building documentation,
an ItemModelTable to easily display Qt ItemModels as Markdown Tables
as well as a model + widget to preview a generated mknodes tree.

"""

from __future__ import annotations

from .itemmodeltable import DependencyTable, ItemModelTable
from .markdownmodel import MarkdownModel
from .markdownwidget import MarkdownWidget
from .prettyqtclasspage import PrettyQtClassPage
from .prettyqtdiagram import PrettyQtDiagram
from .propertytable import PropertyTable
from .widgetscreenshot import WidgetScreenShot


__all__ = [
    "ItemModelTable",
    "DependencyTable",
    "MarkdownModel",
    "MarkdownWidget",
    "PrettyQtClassPage",
    "PrettyQtDiagram",
    "PropertyTable",
    "WidgetScreenShot",
]



from __future__ import annotations

from prettyqt import widgets
from prettyqt.utils import bidict


CHECK_TYPE = bidict(
    not_checkable=widgets.QStyleOptionMenuItem.CheckType.NotCheckable,
    exclusive=widgets.QStyleOptionMenuItem.CheckType.Exclusive,
    non_exclusive=widgets.QStyleOptionMenuItem.CheckType.NonExclusive,
)

MENU_ITEM_TYPE = bidict(
    normal=widgets.QStyleOptionMenuItem.MenuItemType.Normal,
    default_item=widgets.QStyleOptionMenuItem.MenuItemType.DefaultItem,
    separator=widgets.QStyleOptionMenuItem.MenuItemType.Separator,
    sub_menu=widgets.QStyleOptionMenuItem.MenuItemType.SubMenu,
    scroller=widgets.QStyleOptionMenuItem.MenuItemType.Scroller,
    tear_off=widgets.QStyleOptionMenuItem.MenuItemType.TearOff,
    margin=widgets.QStyleOptionMenuItem.MenuItemType.Margin,
    empty_area=widgets.QStyleOptionMenuItem.MenuItemType.EmptyArea,
)


class StyleOptionMenuItem(widgets.StyleOptionMixin, widgets.QStyleOptionMenuItem):
    pass



from __future__ import annotations

from typing import Literal

from prettyqt import core
from prettyqt.utils import bidict


ProcessEventFlagStr = Literal[
    "all", "exclude_user_input", "exclude_socket_notifiers", "wait_for_more"
]

PROCESS_EVENT_FLAGS: bidict[ProcessEventFlagStr, core.QEventLoop.ProcessEventsFlag] = (
    bidict(
        all=core.QEventLoop.ProcessEventsFlag.AllEvents,
        exclude_user_input=core.QEventLoop.ProcessEventsFlag.ExcludeUserInputEvents,
        exclude_socket_notifiers=core.QEventLoop.ProcessEventsFlag.ExcludeSocketNotifiers,
        wait_for_more=core.QEventLoop.ProcessEventsFlag.WaitForMoreEvents,
    )
)


class EventLoop(core.ObjectMixin, core.QEventLoop):
    """Means of entering and leaving an event loop."""

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._executing = False

    def execute(
        self,
        user_input: bool = True,
        socket_notifiers: bool = True,
        wait_for_more: bool = False,
    ) -> int:
        if self._executing:
            raise AssertionError("Eventloop is already running!")
        self._executing = True
        flag = core.QEventLoop.ProcessEventsFlag(0)
        if not user_input:
            flag |= 1
        if not socket_notifiers:
            flag |= 2
        if wait_for_more:
            flag |= 4
        status = self.exec(flag)
        self._executing = False
        return status



from __future__ import annotations

from prettyqt import constants, core
from prettyqt.utils import get_repr


class Collator(core.QCollator):
    """Compares strings according to a localized collation algorithm."""

    def __repr__(self):
        return get_repr(self, self.get_locale())

    def get_locale(self) -> core.Locale:
        return core.Locale(self.locale())

    def set_case_sensitive(self, state: bool):
        """Set case sensitivity.

        Args:
            state: case sensitive

        """
        sensitivity = (
            constants.CaseSensitivity.CaseSensitive
            if state
            else constants.CaseSensitivity.CaseInsensitive
        )
        self.setCaseSensitivity(sensitivity)

    def is_case_sensitive(self) -> bool:
        """Return case sensitivity.

        Returns:
            case sensitivity
        """
        return bool(self.caseSensitivity())

    def get_sort_key(self, string: str) -> core.CollatorSortKey:
        return core.CollatorSortKey(self.sortKey(string))



"""Module containing ItemModels for common data structures.

PrettyQt includes a large amount of ItemModels for Qt-based types,
Python built-in types as well as for different data structures of external libraries.

* All models are proper views on data structures, not populated StandardItemModels.
* Boolean values are always handled via CheckState role, both for editing and displaying.
* Some of the models should be used in conjunction with the
[EditorDelegate](editordelegate.md). That delegate supports editing a large amount of
different datatypes and should be the
preferred choice for most models.
* In general, the models are unstyled (with some few exceptions.
Styling should be done via the extensive [proxy system](proxies.md)
which is baked into PrettyQt.

"""

# BaseModels

from .selectionmixin import SelectionMixin
from .listmixin import ListMixin
from .treemodel import TreeModel
from .columnitemmodel import ColumnItemModel, ColumnTableModel, ColumnItem

# these are deprecated
from .modelmixin import ModelMixin
from .nesteditem import NestedItem
from .nestedmodel import NestedModel

# Python type models

from .regexmatchesmodel import RegexMatchesModel
from .importlibdistributionmodel import ImportlibTreeModel
from .jsonmodel import JsonModel
from .astmodel import AstModel
from .classtreemodel import SubClassTreeModel, ParentClassTreeModel
from .frameinfomodel import FrameInfoModel
from .moduleinfomodel import ModuleInfoModel
from .basefieldsmodel import BaseFieldsModel
from .basedataclassmodel import BaseDataclassModel
from .dataclassmodel import DataClassModel
from .dataclassfieldsmodel import DataClassFieldsModel
from .logrecordmodel import LogRecordModel
from .xmlmodel import BaseXmlModel, XmlModel
from .mappingmodel import MappingModel
from .tupletreemodel import TupleTreeModel
from .pythonobjecttreemodel import PythonObjectTreeModel

# Qt type models

from .actionsmodel import ActionsModel
from .modelindexmodel import ModelIndexModel
from .storageinfomodel import StorageInfoModel
from .shortcutsmodel import ShortcutsModel
from .qobjectpropertiesmodel import QObjectPropertiesModel
from .widgetsdetailsmodel import WidgetsDetailsModel
from .widgethierarchymodel import WidgetHierarchyModel, LayoutHierarchyModel

# Slice Proxies

from .proxies.sliceidentityproxymodel import SliceIdentityProxyModel
from .proxies.slicedisplaytextproxymodel import SliceDisplayTextProxyModel
from .proxies.slicechangeiconsizeproxymodel import SliceChangeIconSizeProxyModel
from .proxies.slicefilterproxymodel import SliceFilterProxyModel
from .proxies.slicecheckableproxymodel import (
    SliceCheckableProxyModel,
    SliceCheckableTreeProxyModel,
)
from .proxies.sliceappearanceproxymodel import SliceAppearanceProxyModel
from .proxies.slicechangeflagsproxymodel import SliceChangeFlagsProxyModel
from .proxies.slicecolorvaluesproxymodel import SliceColorValuesProxyModel
from .proxies.slicecolorcategoriesproxymodel import SliceColorCategoriesProxyModel
from .proxies.slicevaluetransformationproxymodel import SliceValueTransformationProxyModel
from .proxies.slicehighlightcurrentproxymodel import SliceHighlightCurrentProxyModel
from .proxies.slicemaproleproxymodel import SliceMapRoleProxyModel
from .proxies.slicetomarkdownproxymodel import SliceToMarkdownProxyModel

# Reshape proxies

from .proxies.tabletolistproxymodel import TableToListProxyModel
from .proxies.flattentreeproxymodel import FlattenTreeProxyModel
from .proxies.meltproxymodel import MeltProxyModel
from .proxies.columnorderproxymodel import ColumnOrderProxyModel

# Other proxies
from .proxies.fuzzyfilterproxymodel import FuzzyFilterProxyModel
from .proxies.multicolumnfilterproxymodel import MultiColumnFilterProxyModel
from .proxies.subsetfilterproxymodel import SubsetFilterProxyModel
from .proxies.valuefilterproxymodel import ValueFilterProxyModel
from .proxies.rangefilterproxymodel import RangeFilterProxyModel
from .proxies.appearanceproxymodel import AppearanceProxyModel
from .proxies.changeheadersproxymodel import ChangeHeadersProxyModel
from .proxies.columnjoinerproxymodel import ColumnJoinerProxyModel
from .proxies.predicatefilterproxymodel import PredicateFilterProxyModel
from .proxies.highlightmouseproxymodel import HighlightMouseProxyModel

# tooling
from .itemmodelresolver import ItemModelResolver
from .proxies.proxymapper import ProxyMapper
from .proxies.linkedselectionmodel import LinkedSelectionModel

# Models for "external types"
# import importlib.util

# if importlib.util.find_spec("attrs") is not None:
#     from .attrsmodel import AttrsModel
#     from .attrsfieldsmodel import AttrsFieldsModel
# if importlib.util.find_spec("pydantic") is not None:
#     from .pydanticmodel import PydanticModel
#     from .pydanticfieldsmodel import PydanticFieldsModel

__all__ = [
    "SelectionMixin",
    "TreeModel",
    "ImportlibTreeModel",
    "ModelIndexModel",
    "RegexMatchesModel",
    "ColumnItemModel",
    "ColumnTableModel",
    "ListMixin",
    "ModelMixin",
    "ColumnItem",
    "NestedModel",
    "NestedItem",
    "JsonModel",
    "ActionsModel",
    "MappingModel",
    "TupleTreeModel",
    "PythonObjectTreeModel",
    "AstModel",
    "SubClassTreeModel",
    "ParentClassTreeModel",
    "FrameInfoModel",
    "ModuleInfoModel",
    "BaseFieldsModel",
    "BaseDataclassModel",
    "DataClassModel",
    "StorageInfoModel",
    "ShortcutsModel",
    "DataClassFieldsModel",
    "LogRecordModel",
    "BaseXmlModel",
    "XmlModel",
    "QObjectPropertiesModel",
    "WidgetHierarchyModel",
    "LayoutHierarchyModel",
    "WidgetsDetailsModel",
    "SliceChangeIconSizeProxyModel",
    "MultiColumnFilterProxyModel",
    "SliceIdentityProxyModel",
    "SliceDisplayTextProxyModel",
    "SliceFilterProxyModel",
    "FuzzyFilterProxyModel",
    "SubsetFilterProxyModel",
    "SliceValueTransformationProxyModel",
    "ValueFilterProxyModel",
    "RangeFilterProxyModel",
    "SliceCheckableProxyModel",
    "SliceCheckableTreeProxyModel",
    "SliceToMarkdownProxyModel",
    "TableToListProxyModel",
    "FlattenTreeProxyModel",
    "SliceColorCategoriesProxyModel",
    "MeltProxyModel",
    "ColumnOrderProxyModel",
    "SliceAppearanceProxyModel",
    "ChangeHeadersProxyModel",
    "AppearanceProxyModel",
    "ColumnJoinerProxyModel",
    "PredicateFilterProxyModel",
    "SliceChangeFlagsProxyModel",
    "SliceHighlightCurrentProxyModel",
    "SliceMapRoleProxyModel",
    "SliceColorValuesProxyModel",
    "HighlightMouseProxyModel",
    "ProxyMapper",
    "ItemModelResolver",
    "LinkedSelectionModel",
    # external
    "attrsmodel",
    "attrsfieldsmodel",
    "pydanticmodel",
    "pydanticfieldsmodel",
    "fsspectreemodel",
]



from __future__ import annotations

import base64
import os
import pathlib

from typing_extensions import Self

from prettyqt import constants, core, gui
from prettyqt.utils import colors, datatypes, serializemixin


class PixmapMixin(serializemixin.SerializeMixin, gui.PaintDeviceMixin):
    """Off-screen image representation that can be used as a paint device."""

    def __bool__(self):
        return not self.isNull()

    def __eq__(self, other):
        return self.toImage() == other.toImage() if isinstance(other, Pixmap) else False

    def __hash__(self):
        return self.cacheKey()

    @classmethod
    def from_file(cls, path: datatypes.PathType) -> Self:
        path = pathlib.Path(path)
        with path.open(mode="rb") as f:
            data = f.read()
        # Create widget
        pixmap = cls()
        pixmap.loadFromData(core.QByteArray(data))
        return pixmap

    @classmethod
    def from_image(
        cls,
        img: gui.QImage,
        color_preference: constants.ColorPreferenceStr = "auto",
        dithering: constants.DitherPreferenceStr = "diffuse",
        alpha_dithering: constants.AlphaDitherPreferenceStr = "threshold",
        mode_preference: constants.ModePreferenceStr = "auto",
    ) -> Self:
        flag = constants.ImageConversionFlag(0)
        flag |= constants.COLOR_PREFERENCE.inverse[color_preference]
        flag |= constants.DITHER_PREFERENCE.inverse[dithering]
        flag |= constants.ALPHA_DITHER_PREFERENCE.inverse[alpha_dithering]
        flag |= constants.MODE_PREFERENCE.inverse[mode_preference]
        return cls(cls.fromImage(img, flag))

    def save_to_file(
        self, path: os.PathLike, fmt: str = "png", quality: int | None = None
    ) -> bool:
        return self.save(os.fspath(path), fmt, -1 if quality is None else quality)

    def get_size(self) -> core.Size:
        return core.Size(self.size())

    def get_rect(self) -> core.Rect:
        return core.Rect(self.rect())

    def to_image(self) -> gui.Image:
        return gui.Image(self.toImage())

    def rotated(self, rotation: int) -> Pixmap:
        w, h = self.width(), self.height()
        pixmap = self.transformed(gui.Transform().rotate(rotation))
        return pixmap.copy((pixmap.width() - w) // 2, (pixmap.height() - h) // 2, w, h)

    def get_image_data_url(self) -> str:
        """Render the contents of the pixmap as a data URL (RFC-2397).

        Returns:
            datauri : str
        """
        device = core.Buffer()
        assert device.open_file("read_write")
        self.save(device, b"png")
        device.close()
        data = bytes(device.data())
        payload = base64.b64encode(data).decode("ascii")
        return f"data:image/png;base64,{payload}"

    @classmethod
    def create_dot(cls, color: datatypes.ColorType = "black", size: int = 16) -> Self:
        col = colors.get_color(color)
        px = cls(size, size)
        px.fill(constants.GlobalColor.transparent)  # type: ignore
        px_size = px.rect().adjusted(1, 1, -1, -1)
        with gui.Painter(px) as painter:
            painter.use_antialiasing()
            painter.setBrush(col)
            pen_color = gui.Color(15, 15, 15)
            painter.set_pen(color=pen_color, width=1)
            painter.drawEllipse(px_size)
        return px

    @classmethod
    def create_checkerboard_pattern(
        cls, n: int, color_1: datatypes.ColorType, color_2: datatypes.ColorType
    ) -> Self:
        """Construct tileable checkerboard pattern for paint events."""
        # Brush will be an nxn checkerboard pattern
        pat = cls(2 * n, 2 * n)
        bg0 = colors.get_color(color_1)
        bg1 = colors.get_color(color_2)
        with gui.Painter(pat) as p:
            p.setPen(constants.PenStyle.NoPen)
            # Paint a checkerboard pattern for the color to be overlaid on
            p.fillRect(pat.rect(), bg0)
            p.fillRect(0, 0, n, n, bg1)
            p.fillRect(n, n, 2 * n, 2 * n, bg1)
        return pat

    @classmethod
    def create_char(
        cls,
        char: str,
        size: int,
        background: datatypes.ColorType = "black",
        color: datatypes.ColorType = "white",
    ) -> Self:
        pixmap = cls(size, size)
        pixmap.fill(constants.GlobalColor.transparent)
        with gui.Painter(pixmap) as painter:
            painter.setRenderHints(
                painter.RenderHint.Antialiasing
                | painter.RenderHint.TextAntialiasing
                | painter.RenderHint.SmoothPixmapTransform
            )
            painter.set_pen(color=background)
            painter.set_brush(background)
            margin = 1 + size // 16
            text_margin = size // 20
            w = size - 2 * margin
            rect = core.RectF(margin, margin, w, w)
            painter.draw_rounded_rect(rect, 30, 30, relative=True)
            painter.set_pen(color=color)
            with painter.edit_font() as font:  # type: gui.QFont
                font.setPixelSize(size - 2 * margin - 2 * text_margin)
            painter.draw_text(rect, char, alignment="center")
        return pixmap


class Pixmap(PixmapMixin, gui.QPixmap):
    pass


if __name__ == "__main__":
    from prettyqt import gui

    app = gui.app()
    p = Pixmap.create_char("A", 20)
    print(bytes(p))



from __future__ import annotations

import enum

from prettyqt import constants, widgets


SE = widgets.QStyle.SubElement


class CenteredBoxProxyStyle(widgets.QProxyStyle):
    class Roles(enum.IntEnum):
        """Proxy roles."""

        CheckAlignmentRole = constants.USER_ROLE + 3321
        DecorationAlignmentRole = constants.USER_ROLE + 1322

    def subElementRect(self, element, option, widget):
        base_res = super().subElementRect(element, option, widget)
        if element != widgets.QStyle.ControlElement.CE_ItemViewItem:
            return base_res
        item_opt = widgets.QStyleOptionViewItem(option)
        # item_opt = QStyleOptionViewItem()
        # widget.initStyleOption(item_opt)
        match element:
            case SE.SE_ItemViewItemCheckIndicator:
                align_data = (
                    item_opt.index.data(self.Roles.CheckAlignmentRole)
                    if item_opt
                    else None
                )
                if align_data is not None:
                    return widgets.QStyle.alignedRect(
                        item_opt.direction,
                        constants.AlignmentFlag(align_data.value),
                        base_res.size(),
                        item_opt.rect,
                    )
            case SE.SE_ItemViewItemDecoration:
                align_data = (
                    item_opt.index.data(self.Roles.DecorationAlignmentRole)
                    if item_opt
                    else None
                )
                if align_data is not None:
                    return widgets.QStyle.alignedRect(
                        item_opt.direction,
                        constants.AlignmentFlag(align_data.value),
                        base_res.size(),
                        item_opt.rect,
                    )
            case SE.SE_ItemViewItemFocusRect:
                check_align_data = (
                    item_opt.index.data(self.Roles.CheckAlignmentRole)
                    if item_opt
                    else None
                )
                decoration_align_data = (
                    item_opt.index.data(self.Roles.DecorationAlignmentRole)
                    if item_opt
                    else None
                )
                if check_align_data is not None or decoration_align_data is not None:
                    return option.rect
        return base_res


if __name__ == "__main__":
    from prettyqt import gui

    app = widgets.app()
    style = CenteredBoxProxyStyle()
    # app.setStyle(style)
    model = gui.StandardItemModel()
    model.setColumnCount(1)
    model.setRowCount(2)
    pix = gui.QPixmap(20, 20)
    pix.fill(constants.GlobalColor.blue)
    checkable_item = gui.StandardItem()
    checkable_item.setFlags(checkable_item.flags() | constants.IS_CHECKABLE)
    checkable_item.setData(constants.ALIGN_CENTER, style.Roles.CheckAlignmentRole)
    model.setItem(0, 0, checkable_item)
    checkable_item = gui.StandardItem()
    checkable_item.setFlags(checkable_item.flags() | constants.IS_CHECKABLE)
    checkable_item.setData(constants.ALIGN_CENTER, style.Roles.DecorationAlignmentRole)
    checkable_item.setIcon(pix)
    model.setItem(1, 0, checkable_item)
    tv = widgets.TableView()
    tv.set_model(model)
    tv.setStyle(style)
    tv.show()
    app.exec()



from __future__ import annotations

from typing import TYPE_CHECKING

from prettyqt import gui


if TYPE_CHECKING:
    from prettyqt.utils import datatypes


def get_color(color: datatypes.ColorType) -> gui.Color:
    """Get gui.Color instance for given parameter.

    named colors are 'aliceblue', 'antiquewhite', 'aqua', 'aquamarine',
    'azure', 'beige', 'bisque', 'black', 'blanchedalmond', 'blue', 'blueviolet',
    'brown', 'burlywood', 'cadetblue', 'chartreuse', 'chocolate', 'coral',
    'cornflowerblue', 'cornsilk', 'crimson', 'cyan', 'darkblue', 'darkcyan',
    'darkgoldenrod', 'darkgray', 'darkgreen', 'darkgrey', 'darkkhaki',
    'darkmagenta', 'darkolivegreen', 'darkorange', 'darkorchid', 'darkred',
    'darksalmon', 'darkseagreen', 'darkslateblue', 'darkslategray', 'darkslategrey',
    'darkturquoise', 'darkviolet', 'deeppink', 'deepskyblue', 'dimgray', 'dimgrey',
    'dodgerblue', 'firebrick', 'floralwhite', 'forestgreen', 'fuchsia', 'gainsboro',
    'ghostwhite', 'gold', 'goldenrod', 'gray', 'green', 'greenyellow', 'grey',
    'honeydew', 'hotpink', 'indianred', 'indigo', 'ivory', 'khaki', 'lavender',
    'lavenderblush', 'lawngreen', 'lemonchiffon', 'lightblue', 'lightcoral',
    'lightcyan', 'lightgoldenrodyellow', 'lightgray', 'lightgreen', 'lightgrey',
    'lightpink', 'lightsalmon', 'lightseagreen', 'lightskyblue', 'lightslategray',
    'lightslategrey', 'lightsteelblue', 'lightyellow', 'lime', 'limegreen', 'linen',
    'magenta', 'maroon', 'mediumaquamarine', 'mediumblue', 'mediumorchid',
    'mediumpurple', 'mediumseagreen', 'mediumslateblue', 'mediumspringgreen',
    'mediumturquoise', 'mediumvioletred', 'midnightblue', 'mintcream', 'mistyrose',
    'moccasin', 'navajowhite', 'navy', 'oldlace', 'olive', 'olivedrab', 'orange',
    'orangered', 'orchid', 'palegoldenrod', 'palegreen', 'paleturquoise',
    'palevioletred', 'papayawhip', 'peachpuff', 'peru', 'pink', 'plum',
    'powderblue', 'purple', 'red', 'rosybrown', 'royalblue', 'saddlebrown',
    'salmon', 'sandybrown', 'seagreen', 'seashell', 'sienna', 'silver', 'skyblue',
    'slateblue', 'slategray', 'slategrey', 'snow', 'springgreen', 'steelblue',
    'tan', 'teal', 'thistle', 'tomato', 'transparent', 'turquoise', 'violet',
    'wheat', 'white', 'whitesmoke', 'yellow', 'yellowgreen'

    Args:
        color (ColorType): color to create gui.Icon from

    Returns:
        gui.Color: color instance
    """
    match color:
        case list() | tuple():
            return gui.Color(*color)
        case str() if color.endswith("_role"):
            color = color.rstrip("_role")
            palette = gui.Palette()
            return palette.get_color(color)
        case _:
            return gui.Color(color)


def interpolate_text_colors(
    bg: datatypes.ColorType, fg: datatypes.ColorType, n_colors: int
) -> list[gui.Color]:
    bg = get_color(bg)
    fg = get_color(fg)
    pal = []
    m = 35
    hue_base = 90 if bg.hue() == -1 else bg.hue()
    for i in range(n_colors):
        h = hue_base + (360.0 / n_colors * i) % 360
        s = 240.0
        v = max(bg.value(), fg.value()) * 0.85
        if (bg.hue() - m < h < bg.hue() + m) or (fg.hue() - m < h < fg.hue() + m):
            h = ((bg.hue() + fg.hue()) / (i + 1)) % 360
            s = ((bg.saturation() + fg.saturation() + 2 * i) / 2) % 256
            v = ((bg.value() + fg.value() + 2 * i) / 2) % 256
        pal.append(gui.Color.from_hsv(h, s, v))
    return pal



from __future__ import annotations

import contextlib
import os

from typing import Literal

from prettyqt.qt import QtCore
from prettyqt.utils import bidict, datatypes


LockErrorStr = Literal["none", "lock_failed", "permission", "unknown"]

LOCK_ERROR: bidict[LockErrorStr, QtCore.QLockFile.LockError] = bidict(
    none=QtCore.QLockFile.LockError.NoError,
    lock_failed=QtCore.QLockFile.LockError.LockFailedError,
    permission=QtCore.QLockFile.LockError.PermissionError,
    unknown=QtCore.QLockFile.LockError.UnknownError,
)


class LockFile(QtCore.QLockFile):
    """Locking between processes using a file."""

    def __init__(self, path: datatypes.PathType):
        super().__init__(os.fspath(path))

    def get_error(self) -> LockErrorStr:
        return LOCK_ERROR.inverse[self.error()]

    @contextlib.contextmanager
    def lock_file(self):
        self.lock()
        yield self
        self.unlock()



from __future__ import annotations

from .baseipythonwidget import BaseIPythonWidget
from .inprocessipythonwidget import InProcessIPythonWidget
from .outofprocessipythonwidget import OutOfProcessIPythonWidget


__all__ = ["InProcessIPythonWidget", "OutOfProcessIPythonWidget", "BaseIPythonWidget"]



from __future__ import annotations

from prettyqt import widgets


class CommonStyleMixin(widgets.StyleMixin):
    pass


class CommonStyle(CommonStyleMixin, widgets.QCommonStyle):
    """Encapsulates the common Look and Feel of a GUI."""


if __name__ == "__main__":
    style = CommonStyle()



from __future__ import annotations

import contextlib
import logging

from prettyqt.qt import QtCore


LOG_MAP = {
    QtCore.QtMsgType.QtDebugMsg: logging.DEBUG,
    QtCore.QtMsgType.QtInfoMsg: logging.INFO,
    QtCore.QtMsgType.QtWarningMsg: logging.WARNING,
    QtCore.QtMsgType.QtCriticalMsg: logging.ERROR,
    QtCore.QtMsgType.QtFatalMsg: logging.CRITICAL,
    QtCore.QtMsgType.QtSystemMsg: logging.CRITICAL,
}

logger = logging.getLogger(__name__)

# qFormatLogMessage(QtMsgType type, const QMessageLogContext context, const QString str)
# qInstallMessageHandler(QtMessageHandler handler)
# qSetMessagePattern(const QString &pattern)


class MessageHandler:
    def __init__(self, logger: logging.Logger):
        self._logger = logger
        self._previous_handler = None

    def install(self):
        self._previous_handler = QtCore.qInstallMessageHandler(self)

    def uninstall(self):
        if self._previous_handler is None:
            QtCore.qInstallMessageHandler(self._previous_handler)

    def __enter__(self):
        self.install()
        return self

    def __exit__(self, *args):
        self.uninstall()

    def __call__(
        self,
        msgtype: QtCore.QtMsgType,
        context: QtCore.QMessageLogContext,
        message: str,
    ):
        ctx = dict.fromkeys(["category", "file", "function", "line"])
        with contextlib.suppress(UnicodeDecodeError):
            ctx["category"] = context.category
        with contextlib.suppress(UnicodeDecodeError):
            ctx["file"] = context.file
        with contextlib.suppress(UnicodeDecodeError):
            ctx["function"] = context.function
        with contextlib.suppress(UnicodeDecodeError):
            ctx["line"] = context.line

        level = LOG_MAP[msgtype]
        self._logger.log(level, message, extra=ctx)


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    handler = MessageHandler(logger)
    with app.debug_mode():
        app.exec()



from __future__ import annotations

from collections.abc import Callable
import webbrowser

from prettyqt import core, webenginecore, widgets
from prettyqt.qt import QtWebEngineWidgets
from prettyqt.utils import datatypes


# os.environ["QTWEBENGINE_CHROMIUM_FLAGS"] = "--enable-logging --log-level=3"


class WebEngineView(widgets.WidgetMixin, QtWebEngineWidgets.QWebEngineView):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setPage(webenginecore.WebEnginePage(self))

    def set_url(self, url: datatypes.UrlType | datatypes.PathType):
        """Set the url of the WebEngineView.

        Clears the view and loads the URL.

        Args:
            url: URL to set
        """
        url = datatypes.to_url(url)
        self.setUrl(url)

    def get_url(self) -> core.Url:
        return core.Url(self.url())

    def load_url(self, url: datatypes.UrlType | datatypes.PathType):
        """Load the URL.

        Loads the specified url and displays it.

        Note: The view remains the same until enough data has arrived
        to display the new URL.

        Args:
            url: URL to load
        """
        url = datatypes.to_url(url)
        self.load(url)

    def set_zoom(self, zoom: float):
        """Set the zoom factor for the view.

        Valid values are within the range from 0.25 to 5.0. The default factor is 1.0.

        Args:
            zoom: Zoom factor
        """
        self.setZoomFactor(zoom)

    def find_text(
        self,
        string: str,
        backward: bool = False,
        case_sensitive: bool = False,
        callback: Callable[[bool], None] | None = None,
    ):
        """Find text in the current page.

        Finds the specified string, subString, in the page, using the given options.
        The findTextFinished() signal is emitted when a string search is completed.

        To clear the search highlight, just pass an empty string.

        The resultCallback must take a boolean parameter.
        It will be called with a value of true if the subString was found;
        otherwise the callback value will be false.

        Warning: It is guaranteed that the callback is always called,
        but it might be done during page destruction. When WebEnginePage is deleted,
        the callback is triggered with an invalid value and it is not safe to use
        the corresponding QWebEnginePage or QWebEngineView instance inside it.

        Args:
            string: string to search for
            backward: search backwards
            case_sensitive: case-sensitive search
            callback: result callback
        """
        if callback is None:

            def do_nothing(x):
                pass

            callback = do_nothing
        flag = webenginecore.WebEnginePage.FindFlag(0)
        if case_sensitive:
            flag |= webenginecore.WebEnginePage.FindFlag.FindCaseSensitively
        if backward:
            flag |= webenginecore.WebEnginePage.FindFlag.FindBackward
        self.findText(string, flag, callback)

    def get_settings(self) -> webenginecore.WebEngineSettings:
        settings = self.settings()
        return webenginecore.WebEngineSettings(settings)

    def set_setting(
        self,
        setting_name: webenginecore.webenginesettings.WebAttributeStr,
        value: bool,
    ):
        self.get_settings()[setting_name] = value

    def get_setting(
        self, setting_name: webenginecore.webenginesettings.WebAttributeStr
    ) -> bool:
        return self.get_settings()[setting_name]

    @classmethod
    def register_as_browser(cls, tabwidget: widgets.TabWidget):
        class BuiltInBrowser(webbrowser.BaseBrowser):
            def open(self, url: str, new: int = 0, autoraise: bool = True):
                # logger.info(f"opening {url} with builtin browser..")
                webview = cls()
                webview.load_url(url)
                if new == 1:
                    webview.show()
                else:
                    tabwidget.add_tab(webview, url, show=autoraise)

        webbrowser.register("BuiltInBrowser", BuiltInBrowser)

    def last_context_menu_request(
        self,
    ) -> webenginecore.WebEngineContextMenuRequest | None:
        req = self.lastContextMenuRequest()
        return webenginecore.WebEngineContextMenuRequest(req) if req else None


if __name__ == "__main__":
    import pathlib

    from prettyqt import widgets

    app = widgets.app()
    widget = WebEngineView()

    import markdown

    path = pathlib.Path("E:\\dev\\prettyqt\\docs\\index.md")

    extensions = [
        "pymdownx.arithmatex",
        "pymdownx.betterem",
        "pymdownx.caret",
        "pymdownx.critic",
        "pymdownx.details",
        "pymdownx.emoji",
        "pymdownx.inlinehilite",
        "pymdownx.magiclink",
        "pymdownx.mark",
        "pymdownx.smartsymbols",
        "pymdownx.superfences",
        "pymdownx.tasklist",
        "pymdownx.tabbed",
        "pymdownx.tilde",
    ]

    html = markdown.markdown(path.read_text(), extensions=extensions)
    widget.setHtml(html)
    page = widget.page()
    path = pathlib.Path(
        "E:\\dev\\prettyqt\\site\\assets\\stylesheets\\main.26e3688c.min.css"
    )
    page.insert_stylesheet("main.26e3688c.min", path)
    widget.show()
    app.exec()



from __future__ import annotations

from typing import Literal

from prettyqt import core
from prettyqt.qt import QtGui
from prettyqt.utils import bidict, datatypes, get_repr


CursorPositionStr = Literal["cursor_between_characters", "cursor_on_character"]

CURSOR_POSITION: bidict[CursorPositionStr, QtGui.QTextLine.CursorPosition] = bidict(
    cursor_between_characters=QtGui.QTextLine.CursorPosition.CursorBetweenCharacters,
    cursor_on_character=QtGui.QTextLine.CursorPosition.CursorOnCharacter,
)

EdgeStr = Literal["leading", "trailing"]

EDGE: bidict[EdgeStr, QtGui.QTextLine.Edge] = bidict(
    leading=QtGui.QTextLine.Edge.Leading,
    trailing=QtGui.QTextLine.Edge.Trailing,
)


class TextLine(QtGui.QTextLine):
    def __bool__(self):
        return self.isValid()

    def __repr__(self):
        return get_repr(self)

    def __len__(self):
        return self.textLength()

    def get_position(self) -> core.PointF:
        return core.PointF(self.position())

    def set_position(self, point: datatypes.PointType):
        p = core.Point(*point) if isinstance(point, tuple) else point
        self.setPosition(p)

    def cursor_to_x(
        self, cursor_pos: int, edge: EdgeStr | QtGui.QTextLine.Edge = "leading"
    ) -> float:
        return self.cursorToX(cursor_pos, EDGE.get_enum_value(edge))  # type: ignore

    def x_to_cursor(
        self,
        x: float,
        cursor_pos: CursorPositionStr
        | QtGui.QTextLine.CursorPosition = "cursor_between_characters",
    ) -> int:
        return self.xToCursor(x, CURSOR_POSITION.get_enum_value(cursor_pos))


if __name__ == "__main__":
    fmt = TextLine()



from __future__ import annotations

from typing import Literal

from prettyqt import constants, core, gui, widgets
from prettyqt.utils import bidict, datatypes


mod = widgets.QGraphicsView


DragModeStr = Literal["none", "scroll_hand", "rubber_band"]

DRAG_MODE: bidict[DragModeStr, mod.DragMode] = bidict(
    none=mod.DragMode.NoDrag,
    scroll_hand=mod.DragMode.ScrollHandDrag,
    rubber_band=mod.DragMode.RubberBandDrag,
)

CacheModeStr = Literal["none", "background"]

CACHE_MODES: bidict[CacheModeStr, mod.CacheModeFlag] = bidict(
    none=mod.CacheModeFlag.CacheNone,
    background=mod.CacheModeFlag.CacheBackground,
)

OptimizationFlagStr = Literal["dont_save_painter_state", "dont_adjust_for_antialiasing"]

OPTIMIZATION_FLAGS: bidict[OptimizationFlagStr, mod.OptimizationFlag] = bidict(
    # dont_clip_painter=mod.OptimizationFlag.DontClipPainter,
    dont_save_painter_state=mod.OptimizationFlag.DontSavePainterState,
    dont_adjust_for_antialiasing=mod.OptimizationFlag.DontAdjustForAntialiasing,
)

ViewportAnchorStr = Literal["none", "view_center", "under_mouse"]

VIEWPORT_ANCHOR: bidict[ViewportAnchorStr, mod.ViewportAnchor] = bidict(
    none=mod.ViewportAnchor.NoAnchor,
    view_center=mod.ViewportAnchor.AnchorViewCenter,
    under_mouse=mod.ViewportAnchor.AnchorUnderMouse,
)

ViewportUpdateModeStr = Literal["full", "minimal", "smart", "bounding_rect", "none"]

VIEWPORT_UPDATE_MODE: bidict[ViewportUpdateModeStr, mod.ViewportUpdateMode] = bidict(
    full=mod.ViewportUpdateMode.FullViewportUpdate,
    minimal=mod.ViewportUpdateMode.MinimalViewportUpdate,
    smart=mod.ViewportUpdateMode.SmartViewportUpdate,
    bounding_rect=mod.ViewportUpdateMode.BoundingRectViewportUpdate,
    none=mod.ViewportUpdateMode.NoViewportUpdate,
)


class GraphicsViewMixin(widgets.AbstractScrollAreaMixin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not args or not isinstance(args[0], widgets.QGraphicsScene):
            self.setScene(widgets.GraphicsScene())

    def _get_map(self):
        maps = super()._get_map()
        maps |= {
            "alignment": constants.ALIGNMENTS,
            "resizeAnchor": VIEWPORT_ANCHOR,
            "transformationAnchor": VIEWPORT_ANCHOR,
            "cacheMode": CACHE_MODES,
            "dragMode": DRAG_MODE,
            "viewportUpdateMode": VIEWPORT_UPDATE_MODE,
            "rubberBandSelectionMode": constants.ITEM_SELECTION_MODE,
            "renderHints": gui.painter.RENDER_HINTS,
            "optimizationFlags": OPTIMIZATION_FLAGS,
        }
        return maps

    def enable_mousewheel_zoom(self, state: bool = True):
        if state:
            self.viewport().installEventFilter(self)
        else:
            self.viewport().removeEventFilter(self)

    def __getitem__(self, index: int) -> widgets.QGraphicsItem:
        return self.items()[index]

    def eventFilter(self, source, event) -> bool:
        if source is not self.viewport() or event.type() != event.Type.Wheel:
            return super().eventFilter(source, event)
        # Zoom Factor
        zoom_in_factor = 1.25
        zoom_out_factor = 1 / zoom_in_factor

        # Set Anchors
        self.setTransformationAnchor(self.ViewportAnchor.NoAnchor)
        self.setResizeAnchor(self.ViewportAnchor.NoAnchor)

        # Save the scene pos
        old_pos = self.mapToScene(event.position().toPoint())

        # Zoom
        zoom_factor = zoom_in_factor if event.angleDelta().y() > 0 else zoom_out_factor
        self.scale(zoom_factor, zoom_factor)

        # # Get the new position
        new_pos = self.mapToScene(event.position().toPoint())

        # Move scene to old position
        delta = new_pos - old_pos
        self.translate(delta.x(), delta.y())
        return True

    def get_zoom(self) -> float:
        """Return the viewer zoom level.

        Returns:
            float: zoom level.
        """
        transform = self.transform()
        cur_scale = (transform.m11(), transform.m22())
        return float(f"{cur_scale[0] - 1.0:0.2f}")

    def add_item(self, *args):
        return self.scene().addItem(*args)

    def remove_item(self, *args):
        return self.scene().removeItem(*args)

    def get_view_rect(self) -> core.QRect:
        """Return the boundaries of the view in scene coordinates."""
        r = core.QRectF(self.rect())
        return self.viewportTransform().inverted()[0].mapRect(r)

    def get_pixel_size(self):
        """Return vector with length and width of one view pixel in scene coordinates."""
        p0 = core.PointF(0, 0)
        p1 = core.PointF(1, 1)
        tr = self.transform().inverted()[0]
        p01 = tr.map(p0)
        p11 = tr.map(p1)
        return core.PointF(p11 - p01)

    def get_background_brush(self) -> gui.Brush:
        return gui.Brush(self.backgroundBrush())

    def get_foreground_brush(self) -> gui.Brush:
        return gui.Brush(self.foregroundBrush())

    def invalidate_scene(
        self,
        rect: core.QRectF,
        layer: widgets.graphicsscene.SceneLayerStr
        | widgets.QGraphicsScene.SceneLayer = "all",
    ):
        self.invalidateScene(
            rect, widgets.graphicsscene.SCENE_LAYER.get_enum_value(layer)
        )

    def set_transformation_anchor(self, mode: ViewportAnchorStr | mod.ViewportAnchor):
        """Set how the view should position the scene during transformations.

        Args:
            mode: transformation anchor to use
        """
        self.setTransformationAnchor(VIEWPORT_ANCHOR.get_enum_value(mode))

    def get_transformation_anchor(self) -> ViewportAnchorStr:
        """Return current transformation anchor.

        Returns:
            viewport anchor
        """
        return VIEWPORT_ANCHOR.inverse[self.transformationAnchor()]

    def set_transform(self, transform: datatypes.TransformType, combine: bool = False):
        self.setTransform(datatypes.to_transform(transform), combine)

    def set_resize_anchor(self, mode: ViewportAnchorStr | mod.ViewportAnchor):
        """Set how the view should position the scene during resizes.

        Args:
            mode: resize anchor to use
        """
        self.setResizeAnchor(VIEWPORT_ANCHOR.get_enum_value(mode))

    def get_resize_anchor(self) -> ViewportAnchorStr:
        """Return current resize anchor.

        Returns:
            resize anchor
        """
        return VIEWPORT_ANCHOR.inverse[self.resizeAnchor()]

    def set_viewport_update_mode(
        self, mode: ViewportUpdateModeStr | mod.ViewportUpdateMode
    ):
        """Set how the viewport should update its contents.

        Args:
            mode: viewport update mode to use
        """
        self.setViewportUpdateMode(VIEWPORT_UPDATE_MODE.get_enum_value(mode))

    def get_viewport_update_mode(self) -> ViewportUpdateModeStr:
        """Return current viewport update mode.

        Returns:
            viewport update mode
        """
        return VIEWPORT_UPDATE_MODE.inverse[self.viewportUpdateMode()]

    def set_drag_mode(self, mode: DragModeStr | mod.DragMode):
        """Set the behavior for dragging the mouse while the left mouse button is pressed.

        Args:
            mode: drag mode to use
        """
        self.setDragMode(DRAG_MODE.get_enum_value(mode))

    def get_drag_mode(self) -> DragModeStr:
        """Return current drag mode.

        Returns:
            drag mode
        """
        return DRAG_MODE.inverse[self.dragMode()]

    def set_rubberband_selection_mode(
        self, mode: constants.ItemSelectionModeStr | constants.ItemSelectionMode
    ):
        """Set the behavior for selecting items with a rubber band selection rectangle.

        Args:
            mode: rubberband selection mode to use
        """
        self.setRubberBandSelectionMode(
            constants.ITEM_SELECTION_MODE.get_enum_value(mode)
        )

    def get_rubberband_selection_mode(self) -> constants.ItemSelectionModeStr:
        """Return current rubberband selection mode.

        Returns:
            rubberband selection mode
        """
        return constants.ITEM_SELECTION_MODE.inverse[self.rubberBandSelectionMode()]

    def set_cache_mode(self, mode: CacheModeStr | mod.CacheModeFlag):
        """Set the cache mode.

        Args:
            mode: cache mode to use
        """
        self.setCacheMode(CACHE_MODES.get_enum_value(mode))

    def get_cache_mode(self) -> CacheModeStr:
        """Return current cache mode.

        Returns:
            cache mode
        """
        return CACHE_MODES.inverse[self.cacheMode()]

    def set_optimization_flags(self, *items: OptimizationFlagStr):
        flags = OPTIMIZATION_FLAGS.merge_flags(items)
        self.setOptimizationFlags(flags)

    def get_optimization_flags(self) -> list[OptimizationFlagStr]:
        return OPTIMIZATION_FLAGS.get_list(self.optimizationFlags())


class GraphicsView(GraphicsViewMixin, widgets.QGraphicsView):
    """Widget for displaying the contents of a QGraphicsScene."""


if __name__ == "__main__":
    app = widgets.app()
    view = GraphicsView()
    scene = widgets.GraphicsScene()
    scene.add_line(core.Line(0, 0, 200, 100))
    view.setScene(scene)
    view.show()
    app.exec()



from __future__ import annotations

from collections.abc import Iterable, Mapping
from dataclasses import dataclass
from typing import Any

from prettyqt import constants, itemmodels


class NameColumn(itemmodels.ColumnItem):
    name = "Name"
    doc = "Name"

    def get_data(self, item, role):
        match role:
            case constants.DISPLAY_ROLE:
                return item.obj.key


class ValueColumn(itemmodels.ColumnItem):
    name = "Value"
    doc = "Value"
    editable = True

    def get_data(self, item, role):
        match role:
            case constants.DISPLAY_ROLE:
                return repr(item.obj.value)
            case constants.EDIT_ROLE:
                return item.obj.value

    def set_data(self, item, value, role):
        match role:
            case constants.EDIT_ROLE:
                item.obj.value = value
                return True
        return False


class TypeColumn(itemmodels.ColumnItem):
    name = "Type"
    doc = "Type"

    def get_data(self, item, role):
        match role:
            case constants.DISPLAY_ROLE:
                return repr(item.obj.typ)


@dataclass
class JsonItem:
    key: str | int
    value: Any
    typ: type | None = None


class JsonModel(itemmodels.ColumnItemModel):
    COLUMNS = [NameColumn, ValueColumn, TypeColumn]
    SUPPORTS = Mapping

    def __init__(
        self,
        obj: Any,
        show_root: bool = True,
        **kwargs,
    ):
        super().__init__(
            obj=JsonItem(key="", value=obj, typ=type(obj)),
            columns=self.COLUMNS,
            show_root=show_root,
            **kwargs,
        )

    @classmethod
    def supports(cls, instance) -> bool:
        return isinstance(instance, Mapping)

    def _has_children(self, item: JsonModel.TreeItem) -> bool:
        return isinstance(item.obj.value, dict | list | set) and bool(item.obj.value)

    def _fetch_object_children(
        self, item: JsonModel.TreeItem
    ) -> list[JsonModel.TreeItem]:
        """Fetch the children of a Python object.

        Returns: list of treeitem.TreeItems
        """
        # items = []
        match item.obj.value:
            case Mapping():
                return [
                    self.TreeItem(obj=JsonItem(key=k, value=v, typ=type(v)))
                    for k, v in item.obj.value.items()
                ]

            case Iterable() if not isinstance(item.obj.value, str):
                return [
                    self.TreeItem(obj=JsonItem(key=k, value=v, typ=type(v)))
                    for k, v in enumerate(item.obj.value)
                ]
            case _:
                return [
                    self.TreeItem(
                        obj=JsonItem(
                            key="key",
                            value=repr(item.obj.value),
                            typ=type(item.obj.value),
                        )
                    )
                ]


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    dist = [dict(a=2, b={"a": 4, "b": [1, 2, 3], "jkjkjk": "tekjk"}), 6, "jkjk"]
    table = widgets.TreeView()
    model = JsonModel(dist, parent=table)
    table.set_model(model)
    table.proxifier[0].modify(lambda x: x * 2)
    table.proxifier[1].modify(lambda x: x * 4)
    table.setRootIsDecorated(True)
    # table.setSortingEnabled(True)
    table.show()
    app.exec()



from __future__ import annotations

from prettyqt import widgets
from prettyqt.utils import bidict


BUTTON_FEATURES = bidict(
    none=widgets.QStyleOptionButton.ButtonFeature.None_,
    flat=widgets.QStyleOptionButton.ButtonFeature.Flat,
    has_menu=widgets.QStyleOptionButton.ButtonFeature.HasMenu,
    default_button=widgets.QStyleOptionButton.ButtonFeature.DefaultButton,
    auto_default_button=widgets.QStyleOptionButton.ButtonFeature.AutoDefaultButton,
    command_link_button=widgets.QStyleOptionButton.ButtonFeature.CommandLinkButton,
)


class StyleOptionButton(widgets.StyleOptionMixin, widgets.QStyleOptionButton):
    pass



from __future__ import annotations

from prettyqt import core
from prettyqt.qt import QtDesigner


class PyDesignerTaskMenuExtension(
    core.ObjectMixin, QtDesigner.QPyDesignerTaskMenuExtension
):
    def preferredEditAction(self):
        return NotImplemented

    def taskActions(self):
        return NotImplemented


if __name__ == "__main__":
    obj = core.Object()
    ext = PyDesignerTaskMenuExtension(obj)
    new = dir(PyDesignerTaskMenuExtension(obj))
    bie = dir(core.Object())
    c = [x for x in new if x not in bie]



from __future__ import annotations

from prettyqt import core


class StringListModelMixin(core.AbstractListModelMixin):
    pass


class StringListModel(StringListModelMixin, core.QStringListModel):
    """Model that supplies strings to views."""

    def __repr__(self):
        return f"{type(self).__name__}: ({self.rowCount()})"

    def columnCount(self, parent: core.ModelIndex | None = None) -> int:
        # columnCount is private for StringListModel, but we need it
        # to avoid workarounds (for example in our Slice proxies).
        # So lets just return 1.
        return 1

    @classmethod
    def supports(cls, instance) -> bool:
        match instance:
            case (str(), *_):
                return True
            case _:
                return False



from __future__ import annotations

from typing import Literal

from prettyqt import gui
from prettyqt.utils import bidict, get_repr


ModeStr = Literal["standard", "full_page"]

MODES: bidict[ModeStr, gui.QPageLayout.Mode] = bidict(
    standard=gui.QPageLayout.Mode.StandardMode,
    full_page=gui.QPageLayout.Mode.FullPageMode,
)

OrientationStr = Literal["portrait", "landscape"]

ORIENTATIONS: bidict[OrientationStr, gui.QPageLayout.Orientation] = bidict(
    portrait=gui.QPageLayout.Orientation.Portrait,
    landscape=gui.QPageLayout.Orientation.Landscape,
)

UnitStr = Literal["millimeter", "point", "inch", "pica", "didot", "cicero"]

UNITS: bidict[UnitStr, gui.QPageLayout.Unit] = bidict(
    millimeter=gui.QPageLayout.Unit.Millimeter,
    point=gui.QPageLayout.Unit.Point,
    inch=gui.QPageLayout.Unit.Inch,
    pica=gui.QPageLayout.Unit.Pica,
    didot=gui.QPageLayout.Unit.Didot,
    cicero=gui.QPageLayout.Unit.Cicero,
)


class PageLayout(gui.QPageLayout):
    """Describes the size, orientation and margins of a page."""

    def __repr__(self):
        return get_repr(self)

    def set_units(self, unit: UnitStr | gui.QPageLayout.Unit):
        """Set unit.

        Args:
            unit: unit
        """
        self.setUnits(UNITS.get_enum_value(unit))

    def get_units(self) -> UnitStr:
        """Get the current unit.

        Returns:
            unit
        """
        return UNITS.inverse[self.units()]

    def set_mode(self, mode: ModeStr | gui.QPageLayout.Mode):
        """Set mode.

        Args:
            mode: mode
        """
        self.setMode(MODES.get_enum_value(mode))

    def get_mode(self) -> ModeStr:
        """Get the current mode.

        Returns:
            mode
        """
        return MODES.inverse[self.mode()]

    def set_orientation(self, orientation: OrientationStr | gui.QPageLayout.Orientation):
        """Set orientation.

        Args:
            orientation: orientation
        """
        self.setOrientation(ORIENTATIONS.get_enum_value(orientation))

    def get_orientation(self) -> OrientationStr:
        """Get the current orientation.

        Returns:
            orientation
        """
        return ORIENTATIONS.inverse[self.orientation()]

    def get_page_size(self) -> gui.PageSize:
        return gui.PageSize(self.pageSize())


if __name__ == "__main__":
    movie = PageLayout()



from __future__ import annotations

from prettyqt import core, eventfilters


class HoverIconEventFilter(eventfilters.BaseEventFilter):
    def __init__(self, normal, hover, pressed=None, parent=None):
        super().__init__(parent)
        self.normal = normal
        self.hover = hover
        self.pressed = pressed

    def eventFilter(self, obj, event: core.Event) -> bool:
        # if obj is not self.parent():
        #     return super().eventFilter(obj, event)
        match event.type():
            case core.Event.Type.Enter | core.Event.Type.MouseButtonRelease if self.hover:
                obj.set_icon(self.hover)
            case core.Event.Type.Leave:
                obj.set_icon(self.normal)
            case core.Event.Type.MouseButtonPress if self.pressed:
                obj.set_icon(self.pressed)
        return super().eventFilter(obj, event)


if __name__ == "__main__":
    from prettyqt import widgets

    app = widgets.app()
    widget = widgets.PushButton()
    widget.set_icon("mdi.folder")
    test = HoverIconEventFilter("mdi.folder", "mdi.folder-outline", "mdi.timer", widget)
    widget.installEventFilter(test)
    widget.show()
    app.exec()



from __future__ import annotations

from prettyqt import core, location
from prettyqt.utils import datatypes


class PlaceSupplier(location.QPlaceSupplier):
    def __bool__(self):
        return not self.isEmpty()

    def get_icon(self) -> location.PlaceIcon | None:
        icon = self.icon()
        return None if icon.isEmpty() else location.PlaceIcon(icon)

    def set_url(self, url: datatypes.UrlType):
        url = core.Url(url)
        self.setUrl(url)

    def get_url(self) -> core.Url:
        return core.Url(self.url())


if __name__ == "__main__":
    supplier = PlaceSupplier()
    print(bool(supplier))



from __future__ import annotations

import logging

from prettyqt import constants, core
from prettyqt.utils import baseresolver


logger = logging.getLogger(__name__)


class ItemModelResolver(baseresolver.BaseResolver):
    """Allows globbing ItemModels."""

    def __init__(
        self,
        model: core.QAbstractItemModel,
        path_role: constants.ItemDataRole = constants.DISPLAY_ROLE,
        ignore_case: bool = False,
    ):
        """Resolve any `ModelIndex` paths using role `path_role`.

        Arguments:
            model: Model to search
            path_role: ItemDataRole to be used for resolving
            ignore_case: Enable case insensisitve handling.
        """
        super().__init__(ignore_case=ignore_case)
        self.model = model
        self.path_role = path_role
        self.fetch_more = True

    def get_parent(self, node):
        return node.parent()

    def get_children(self, node):
        if self.fetch_more:
            while self.model.canFetchMore(node):
                self.model.fetchMore(node)
        return [self.model.index(i, 0, node) for i in range(self.model.rowCount(node))]

    def get_root(self, node):
        return core.ModelIndex()

    def get_attribute(self, node):
        return str(node.data(self.path_role))

    def get(self, path: str, root_node=None):
        path = f"/None{path}"
        return super().get(path, root_node or core.ModelIndex())

    def glob(self, path: str, root_node=None):
        path = f"/None{path}"
        return super().glob(path, root_node or core.ModelIndex())


if __name__ == "__main__":
    from prettyqt import itemmodels, widgets

    app = widgets.app()
    mapping = {"a": "test", ("a", "b"): "test", ("a", "b", "c"): "test"}
    model = itemmodels.TupleTreeModel(mapping)
    resolver = ItemModelResolver(model)
    indexes = resolver.glob("/root/test")
    print(indexes)
    # view = widgets.TreeView()
    # view.setRootIsDecorated(True)
    # view.set_model(model)
    # view.show()
    # with app.debug_mode():
    #     app.exec()



from __future__ import annotations

import logging

from typing import Literal

from prettyqt.qt import QtTextToSpeech
from prettyqt.utils import bidict


logger = logging.getLogger()

AGE = bidict(
    child=QtTextToSpeech.QVoice.Age.Child,
    teenager=QtTextToSpeech.QVoice.Age.Teenager,
    adult=QtTextToSpeech.QVoice.Age.Adult,
    senior=QtTextToSpeech.QVoice.Age.Senior,
    other=QtTextToSpeech.QVoice.Age.Other,
)

AgeStr = Literal["child", "teenager", "adult", "senior", "other"]

GENDER = bidict(
    male=QtTextToSpeech.QVoice.Gender.Male,
    female=QtTextToSpeech.QVoice.Gender.Female,
    unknown=QtTextToSpeech.QVoice.Gender.Unknown,
)

GenderStr = Literal["male", "female", "unknown"]


class Voice(QtTextToSpeech.QVoice):
    def get_age(self) -> AgeStr:
        return AGE.inverse[self.age()]

    def get_gender(self) -> GenderStr:
        return GENDER.inverse[self.gender()]


if __name__ == "__main__":
    val = Voice()



from __future__ import annotations

from typing import Literal

from prettyqt import charts, constants, core, gui, widgets
from prettyqt.utils import bidict


RubberBandStr = Literal["none", "vertical", "horizontal", "rectangle"]

RUBBER_BAND: bidict[RubberBandStr, charts.QChartView.RubberBand] = bidict(
    none=charts.QChartView.RubberBand.NoRubberBand,
    vertical=charts.QChartView.RubberBand.VerticalRubberBand,
    horizontal=charts.QChartView.RubberBand.HorizontalRubberBand,
    rectangle=charts.QChartView.RubberBand.RectangleRubberBand,
)

ZOOM_IN_FACTOR = 1.1
ZOOM_OUT_FACTOR = 1.0 / ZOOM_IN_FACTOR
SCROLL_STEP_SIZE = 10


class ChartView(widgets.GraphicsViewMixin, charts.QChartView):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not args or not isinstance(args[0], charts.QChart):
            self.setChart(charts.Chart())
        self.setRenderHint(gui.Painter.RenderHint.Antialiasing)
        self.set_rubber_band("rectangle")
        # self.setDragMode(self.RubberBandDrag)

    def keyPressEvent(self, event: gui.QKeyEvent):
        """Handle keypress events to allow navigation via keyboard."""
        match event.key():
            case constants.Key.Key_Escape:
                self.chart().zoomReset()
            case constants.Key.Key_Plus:
                self.chart().zoom_by_factor(ZOOM_IN_FACTOR)
            case constants.Key.Key_Minus:
                self.chart().zoom_by_factor(ZOOM_OUT_FACTOR)
            case constants.Key.Key_Left:
                self.chart().scroll(-SCROLL_STEP_SIZE, 0)
            case constants.Key.Key_Right:
                self.chart().scroll(SCROLL_STEP_SIZE, 0)
            case constants.Key.Key_Up:
                self.chart().scroll(0, SCROLL_STEP_SIZE)
            case constants.Key.Key_Down:
                self.chart().scroll(0, -SCROLL_STEP_SIZE)
            case constants.Key.Key_0:
                self.chart().apply_nice_numbers()
            case _:
                super().keyPressEvent(event)
                return
        event.accept()

    def wheelEvent(self, event: gui.QWheelEvent):
        """Handle wheel event for zooming."""
        fct = ZOOM_IN_FACTOR if event.angleDelta().y() > 0 else ZOOM_OUT_FACTOR
        self.chart().zoom_by_factor(fct)
        event.accept()

    def mouseReleaseEvent(self, event: gui.QMouseEvent):
        """Override to allow dragging the chart."""
        if event.button() == constants.MouseButton.RightButton:
            widgets.Application.restoreOverrideCursor()
            event.accept()
            return
        super().mouseReleaseEvent(event)

    def mousePressEvent(self, event: gui.QMouseEvent):
        """Override to allow dragging the chart."""
        if event.button() == constants.MouseButton.RightButton:
            widgets.Application.set_override_cursor("size_all")
            self.last_mouse_pos = event.position()
            event.accept()

        super().mousePressEvent(event)

    def mouseMoveEvent(self, event: gui.QMouseEvent):
        """Override to allow dragging the chart."""
        # pan the chart with a middle mouse drag
        if event.buttons() & constants.MouseButton.RightButton:  # type: ignore
            if not self.last_mouse_pos:
                return
            pos_diff = event.position() - self.last_mouse_pos
            self.chart().scroll(-pos_diff.x(), pos_diff.y())

            self.last_mouse_pos = event.position()
            event.accept()

        super().mouseMoveEvent(event)

    @core.Slot()
    def save_as_image(self):
        """Let user choose folder and save chart as an image file."""
        dlg = widgets.FileDialog(mode="save", caption="Save image")
        filters = {"Bmp files": [".bmp"], "Jpeg files": [".jpg"], "Png files": [".png"]}
        dlg.set_extension_filter(filters)
        filename = dlg.open_file()
        if not filename:
            return
        self.chart().show_legend()
        image = self.get_image()
        image.save(str(filename[0]))
        self.chart().hide_legend()

    def set_rubber_band(self, typ: RubberBandStr | charts.QChartView.RubberBand):
        """Set the rubber band type.

        Args:
            typ: rubber band type
        """
        self.setRubberBand(RUBBER_BAND.get_enum_value(typ))

    def get_rubber_band(self) -> RubberBandStr:
        """Return current rubber band type.

        Returns:
            Rubber band type
        """
        return RUBBER_BAND.inverse[self.rubberBand()]

    # def hide_legend(self):
    #     self.chart().hide_legend()

    # def show_legend(self):
    #     self.chart().show_legend()

    # def set_legend_alignment(self, alignment: str):
    #     if alignment not in constants.SIDES:
    #         raise ValueError(f"{alignment!r} not a valid alignment.")
    #     self.chart().legend().setAlignment(constants.SIDES[alignment])


if __name__ == "__main__":
    app = widgets.app()
    widget = ChartView()
    widget.show()
    app.exec()



from __future__ import annotations

import enum
import logging

from prettyqt import constants, core, gui
from prettyqt.utils import colors, datatypes, fuzzy


logger = logging.getLogger(__name__)


class FuzzyFilterProxyModel(core.SortFilterProxyModel):
    """A FilterProxyModel which sorts the results based on a matching score.

    The matching score evaluates metrics like upper/lower casing, distance of
    filter character to match character and some more.
    Best matches are shown at the top.

    The matching score is exposed as a custom UserRole
    (FuzzyFilterProxyModel.Roles.SortRole)
    The proxymodel replaces the text from the display role in the given filter column
    with HTML code in order to color the letter matches. A backup from the original text
    is made available in the BackupRole. Based on the original text, the proxy calculates
    a score for the match and makes it available via the SortRole.
    To display the html code properly, a HtmlItemDelegate is needed.

    ### Example

    ```py
    proxy = itemmodels.FuzzyFilterProxyModel()
    proxy.set_source_model(model)
    table.set_model(proxy)
    table.show()
    ```
    """

    ID = "fuzzy"

    class Roles(enum.IntEnum):
        """Addional roles."""

        BackupRole = constants.USER_ROLE + 65
        SortRole = constants.SORT_ROLE

    def __init__(self, *args, **kwargs):
        super().__init__(*args, filter_mode="fuzzy", **kwargs)
        self._search_term = ""
        self._match_color: gui.QColor = gui.Color("blue")
        self.sort(0, constants.DESCENDING)

    def set_match_color(self, color: datatypes.ColorType | None):
        self._match_color = colors.get_color(color) if color else gui.QColor()

    def get_match_color(self) -> gui.QColor:
        return self._match_color

    def lessThan(self, left: core.ModelIndex, right: core.ModelIndex):
        if not self._search_term:
            return super().lessThan(left, right)
        if left.data() is None or right.data() is None:
            return True
        # since fuzzy scores are cached, it should be fine to do this here.
        left_data = fuzzy.fuzzy_match(self._search_term, str(left.data()))
        right_data = fuzzy.fuzzy_match(self._search_term, str(right.data()))

        return left_data < right_data

    def set_search_term(self, search_term: str):
        self._search_term = search_term
        super().set_search_term(search_term)
        self.invalidate()

    def get_search_term(self):
        return self._search_term

    def data(
        self,
        index: core.ModelIndex,
        role: constants.ItemDataRole = constants.DISPLAY_ROLE,
    ):
        if not index.isValid():
            return None
        filter_column = self.filterKeyColumn()
        match role, index.column():
            case constants.DISPLAY_ROLE, _ if index.column() == filter_column:
                label = super().data(index, constants.DISPLAY_ROLE)
                return (
                    fuzzy.color_text(
                        self._search_term,
                        str(label),
                        self._match_color.name(),
                        self.is_filter_case_sensitive(),
                    )
                    if self._search_term and self._match_color.isValid() and label
                    else label
                )
            # case constants.DISPLAY_ROLE, 1:
            #     idx = self.index(index.row(), filter_column)
            #     label = super().data(idx, constants.DISPLAY_ROLE)
            #     if label is None:
            #         return None
            #     result = fuzzy.fuzzy_match(self._search_term, str(label))
            #     return str(result[1])
            case self.Roles.BackupRole, _:
                return super().data(index, constants.DISPLAY_ROLE)
            case self.Roles.SortRole, _:
                idx = self.index(index.row(), filter_column)
                label = super().data(idx, constants.DISPLAY_ROLE)
                result = fuzzy.fuzzy_match(self._search_term, label)
                return result[1]
            case _, _:
                return super().data(index, role)

    search_term = core.Property(
        str,
        get_search_term,
        set_search_term,
        doc="Current search term",
    )
    match_color = core.Property(
        gui.QColor,
        get_match_color,
        set_match_color,
        doc="Color to use for match coloring",
    )


if __name__ == "__main__":
    import random
    import string

    from prettyqt import custom_widgets, widgets

    app = widgets.app()
    window = widgets.MainWindow()
    pal = custom_widgets.CommandPalette()
    actions = [
        gui.Action(
            text="super duper action",
            shortcut="Ctrl+A",
            tool_tip="some Tooltip text",
            icon="mdi.folder",
            triggered=lambda: print("test"),
        ),
        gui.Action(
            text="this is an action",
            shortcut="Ctrl+B",
            tool_tip="Tooltip",
            icon="mdi.folder-outline",
            checked=True,
            checkable=True,
        ),
        gui.Action(
            text="another one P",
            shortcut="Ctrl+Alt+A",
            tool_tip="Some longer tooltiPpp",
            icon="mdi.folder",
        ),
        gui.Action(
            text="another onpe P",
            shortcut="Ctrl+Alt+A",
            tool_tip="Some longer tooltiPpp",
            icon="mdi.folder",
        ),
        gui.Action(text="a", shortcut="Ctrl+A", tool_tip="Tooltip", icon="mdi.folder"),
    ]
    pal.populate_from_widget(window)
    pal.add_actions(actions)
    for _ in range(500):
        label = "".join(random.choices(string.ascii_uppercase, k=10))
        pal.add_actions([gui.Action(text=label)])
    pal.show()
    with app.debug_mode():
        app.exec()



from __future__ import annotations

import pathlib

from typing import Literal

from prettyqt.qt import QtCore
from prettyqt.utils import bidict


StandardLocationStr = Literal[
    "desktop",
    "documents",
    "fonts",
    "applications",
    "music",
    "movies",
    "pictures",
    "temp",
    "home",
    "cache",
    "generic_cache",
    "generic_data",
    "runtime",
    "config",
    "download",
    "generic_config",
    "app_data",
    "app_local_data",
    "app_config",
]

STANDARD_LOCATION: bidict[StandardLocationStr, QtCore.QStandardPaths.StandardLocation] = (
    bidict(
        desktop=QtCore.QStandardPaths.StandardLocation.DesktopLocation,
        documents=QtCore.QStandardPaths.StandardLocation.DocumentsLocation,
        fonts=QtCore.QStandardPaths.StandardLocation.FontsLocation,
        applications=QtCore.QStandardPaths.StandardLocation.ApplicationsLocation,
        music=QtCore.QStandardPaths.StandardLocation.MusicLocation,
        movies=QtCore.QStandardPaths.StandardLocation.MoviesLocation,
        pictures=QtCore.QStandardPaths.StandardLocation.PicturesLocation,
        temp=QtCore.QStandardPaths.StandardLocation.TempLocation,
        home=QtCore.QStandardPaths.StandardLocation.HomeLocation,
        cache=QtCore.QStandardPaths.StandardLocation.CacheLocation,
        generic_cache=QtCore.QStandardPaths.StandardLocation.GenericCacheLocation,
        generic_data=QtCore.QStandardPaths.StandardLocation.GenericDataLocation,
        runtime=QtCore.QStandardPaths.StandardLocation.RuntimeLocation,
        config=QtCore.QStandardPaths.StandardLocation.ConfigLocation,
        download=QtCore.QStandardPaths.StandardLocation.DownloadLocation,
        generic_config=QtCore.QStandardPaths.StandardLocation.GenericConfigLocation,
        app_data=QtCore.QStandardPaths.StandardLocation.AppDataLocation,
        app_local_data=QtCore.QStandardPaths.StandardLocation.AppLocalDataLocation,
        app_config=QtCore.QStandardPaths.StandardLocation.AppConfigLocation,
    )
)


class StandardPaths(QtCore.QStandardPaths):
    """Methods for accessing standard paths."""

    @classmethod
    def get_display_name(cls, location: StandardLocationStr) -> str:
        return cls.displayName(STANDARD_LOCATION[location])

    @classmethod
    def get_writable_location(cls, location: StandardLocationStr) -> pathlib.Path | None:
        path = cls.writableLocation(STANDARD_LOCATION[location])
        return pathlib.Path(path) if path else None

    @classmethod
    def get_standard_locations(cls, location: StandardLocationStr) -> list[pathlib.Path]:
        paths = cls.standardLocations(STANDARD_LOCATION[location])
        return [pathlib.Path(p) for p in paths]
